class TestLibrary:
    # Does the library install in scope? Are the objects in scope?
    def test_import(self):
        from dreimac import CircularCoords, ToroidalCoords, ProjectiveCoords, PartUnity, GeometryExamples, CircleMapUtils, PlotUtils
        assert True




import numpy as np
import scipy
from numba import jit
import scipy.sparse as sparse
from scipy.sparse.linalg import lsqr
from scipy.optimize import LinearConstraint, milp
from .utils import PartUnity, CircleMapUtils, CohomologyUtils
from .emcoords import EMCoords
from .combinatorial import (
    combinatorial_number_system_table,
    combinatorial_number_system_d1_forward,
)


class ToroidalCoords(EMCoords):
    """
    Object that performs sparse toroidal coordinates via persistent cohomology as in
    (L. Scoccola, H. Gakhar, J. Bush, N. Schonsheck, T. Rask, L. Zhou, J. Perea 2022)

    Parameters
    ----------
    X: ndarray(N, d)
        A point cloud with N points in d dimensions
    n_landmarks: int
        Number of landmarks to use
    distance_matrix: boolean
        If true, treat X as a distance matrix instead of a point cloud
    prime : int
        Field coefficient with which to compute rips on landmarks
    maxdim : int
        Maximum dimension of homology.  Only dimension 1 is needed for circular coordinates,
        but it may be of interest to see other dimensions (e.g. for a torus)
    """

    def __init__(
        self, X, n_landmarks, distance_matrix=False, prime=41, maxdim=1, verbose=False
    ):
        EMCoords.__init__(self, X, n_landmarks, distance_matrix, prime, maxdim, verbose)
        self.type_ = "toroidal"
        simplicial_complex_dimension = 2
        self.cns_lookup_table_ = combinatorial_number_system_table(
            n_landmarks, simplicial_complex_dimension
        )

    def get_coordinates(
        self,
        perc=0.5,
        cocycle_idxs=[0],
        partunity_fn=PartUnity.linear,
        standard_range=True,
        check_cocycle_condition=True,
    ):
        """
        Get toroidal coordinates.

        Parameters
        ----------
        perc : float
            Percent coverage. Must be between 0 and 1.
        cocycle_idx : integer
            Integer representing the index of the persistent cohomology class
            used to construct the Eilenberg-MacLane coordinate. Persistent cohomology
            classes are ordered by persistence, from largest to smallest.
        partunity_fn : (dist_land_data, r_cover) -> phi
            A function from the distances of each landmark to a bump function
        standard_range : bool
            Whether to use the parameter perc to choose a filtration parameter that guarantees
            that the selected cohomology class represents a non-trivial class in the Cech complex.
        check_cocycle_condition : bool
            Whether to check, and fix if necessary, that the integer cocycle constructed
            using finite field coefficients satisfies the cocycle condition.

        Returns
        -------
        thetas : ndarray(n, N)
            List of circular coordinates, with n the length of cocycle_idxs

        """

        # get representative cocycles and the intersection of their supports
        homological_dimension = 1
        cohomdeaths, cohombirths, cocycles = zip(
            *[
                self.get_representative_cocycle(cohomology_class, homological_dimension)
                for cohomology_class in cocycle_idxs
            ]
        )

        cohomdeath_rips = max(cohomdeaths)
        cohombirth_rips = min(cohombirths)

        if cohomdeath_rips >= cohombirth_rips:
            raise Exception(
                "\
                The supports of the chosen persistent cohomology classes do not intersect"
            )

        # lift to integer cocycles
        integer_cocycles = [
            CohomologyUtils.lift_to_integer_cocycle(cocycle, prime=self.prime_)
            for cocycle in cocycles
        ]

        # determine radius for balls
        r_cover, rips_threshold = EMCoords.get_cover_radius(
            self, perc, cohomdeath_rips, cohombirth_rips, standard_range
        )

        # compute partition of unity and choose a cover element for each data point
        varphi, ball_indx = EMCoords.get_covering_partition(self, r_cover, partunity_fn)

        # compute boundary matrix
        dist_land_land = self.dist_land_land_
        delta0 = CohomologyUtils.make_delta0(
            dist_land_land, rips_threshold, self.cns_lookup_table_
        )

        # go from sparse to dense representation of cocycles
        integer_cocycles_as_vectors = [
            CohomologyUtils.sparse_cocycle_to_vector(
                sparse_cocycle, self.cns_lookup_table_, self.n_landmarks_, int
            )
            for sparse_cocycle in integer_cocycles
        ]

        if check_cocycle_condition:
            delta1 = None
            fixed_cocycles = []

            for class_idx, cocycle_as_vector in enumerate(integer_cocycles_as_vectors):
                is_cocycle, _ = _is_one_cocycle(
                    cocycle_as_vector,
                    dist_land_land,
                    rips_threshold,
                    self.cns_lookup_table_,
                )
                if is_cocycle:
                    fixed_cocycles.append(cocycle_as_vector)
                else:
                    delta1 = (
                        delta1
                        if delta1
                        else CohomologyUtils.make_delta1_compact(
                            dist_land_land, rips_threshold, self.cns_lookup_table_
                        )
                    )

                    d1cocycle = delta1 @ cocycle_as_vector.T

                    y = d1cocycle // self.prime_

                    constraints = LinearConstraint(delta1, y, y)
                    n_edges = delta1.shape[1]
                    objective = np.zeros((n_edges))
                    integrality = np.ones((n_edges))

                    optimizer_solution = milp(
                        objective,
                        integrality=integrality,
                        constraints=constraints,
                    )

                    if not optimizer_solution["success"]:
                        raise Exception(
                            "The cohomology class at index "
                            + str(class_idx)
                            + " does not have an integral lift."
                        )
                    else:
                        solution = optimizer_solution["x"]
                        new_cocycle_as_vector = (
                            cocycle_as_vector
                            - self.prime_ * np.array(np.rint(solution), dtype=int)
                        )

                        fixed_cocycles.append(new_cocycle_as_vector)
            integer_cocycles_as_vectors = fixed_cocycles

        # compute harmonic representatives of cocycles and their circle-valued integrals
        inner_product_matrix, sqrt_inner_product_matrix = _make_inner_product(
            dist_land_land, rips_threshold, self.cns_lookup_table_
        )

        harm_reps_and_integrals = [
            _integrate_harmonic_representative(
                cocycle, delta0, sqrt_inner_product_matrix
            )
            for cocycle in integer_cocycles_as_vectors
        ]
        harm_reps, _ = zip(*harm_reps_and_integrals)

        # compute circular coordinates on data points
        circ_coords = [
            _sparse_integrate(
                harm_rep,
                integral,
                varphi,
                ball_indx,
                dist_land_land,
                rips_threshold,
                self.cns_lookup_table_,
            )
            for harm_rep, integral in harm_reps_and_integrals
        ]

        # if more than one cohomology class was selected
        if len(integer_cocycles) > 1:
            # compute gram matrix with inner products between harmonic representative cocycles
            gram_mat = _gram_matrix(harm_reps, inner_product_matrix)

            # perform lattice reduction on the compute circular coordinates
            # using the gram matrix of the harmonic representative cocycles
            (
                circ_coords,
                change_basis,
                decorrelated_vectors,
            ) = _reduce_circular_coordinates(circ_coords, gram_mat)

            self.original_gram_matrix_ = gram_mat
            self.gram_matrix_ = decorrelated_vectors @ decorrelated_vectors.T
            self.change_basis_ = change_basis

        return circ_coords


def _integrate_harmonic_representative(cocycle, boundary_matrix, sqrt_inner_product):
    integral = lsqr(
        sqrt_inner_product @ boundary_matrix, sqrt_inner_product @ cocycle.T
    )[0]
    harm_rep = cocycle - boundary_matrix.dot(integral)
    return harm_rep, integral


def _make_inner_product(dist_mat, threshold, lookup_table):
    n_points = dist_mat.shape[0]
    n_edges = (n_points * (n_points - 1)) // 2

    max_n_entries = n_edges
    rows = np.empty((max_n_entries,), dtype=int)
    columns = np.empty((max_n_entries,), dtype=int)
    values = np.empty((max_n_entries,), dtype=float)

    @jit(fastmath=True, nopython=True)
    def _make_inner_product_get_row_columns_values(
        dist_mat: np.ndarray,
        threshold: float,
        lookup_table: np.ndarray,
        n_points: int,
        rows: np.ndarray,
        columns: np.ndarray,
        values: np.ndarray,
    ):
        n_entries = 0
        for i in range(n_points):
            for j in range(i + 1, n_points):
                if dist_mat[i, j] < threshold:
                    row_index = combinatorial_number_system_d1_forward(
                        i, j, lookup_table
                    )
                    rows[n_entries] = row_index
                    columns[n_entries] = row_index
                    values[n_entries] = 1
                    n_entries += 1
        return n_entries

    n_entries = _make_inner_product_get_row_columns_values(
        dist_mat, threshold, lookup_table, n_points, rows, columns, values
    )

    W = sparse.csr_array(
        (values[:n_entries], (rows[:n_entries], columns[:n_entries])),
        shape=(n_edges, n_edges),
    )

    WSqrt = W.copy()

    return W, WSqrt


def _sparse_integrate(
    harm_rep,
    integral,
    part_unity,
    membership_function,
    dist_mat,
    threshold,
    lookup_table,
):
    n_points = integral.shape[0]
    theta_matrix = np.zeros((n_points, n_points))

    @jit(fastmath=True, nopython=True)
    def _cocycle_to_matrix(
        dist_mat: np.ndarray,
        threshold: float,
        lookup_table: np.ndarray,
        n_points: int,
        theta_matrix: np.ndarray,
    ):
        for i in range(n_points):
            for j in range(i + 1, n_points):
                if dist_mat[i, j] < threshold:
                    index = combinatorial_number_system_d1_forward(i, j, lookup_table)
                    theta_matrix[i, j] = harm_rep[index]

    _cocycle_to_matrix(dist_mat, threshold, lookup_table, n_points, theta_matrix)

    class_map = integral[membership_function].copy()
    for i in range(class_map.shape[0]):
        class_map[i] += theta_matrix[membership_function[i], :].dot(part_unity[:, i])
    return np.mod(2 * np.pi * class_map, 2 * np.pi)


@jit(fastmath=True, nopython=True)
def _is_one_cocycle(
    cochain: np.ndarray,
    dist_mat: np.ndarray,
    threshold: float,
    lookup_table: np.ndarray,
):
    is_cocycle = True
    first_failure = np.inf
    n_points = dist_mat.shape[0]
    for i in range(n_points):
        for j in range(i + 1, n_points):
            if dist_mat[i, j] < threshold:
                for k in range(j + 1, n_points):
                    if (
                        dist_mat[i, k] < threshold
                        and dist_mat[j, k] < threshold
                    ):
                        index_ij = combinatorial_number_system_d1_forward(
                            i, j, lookup_table
                        )
                        index_jk = combinatorial_number_system_d1_forward(
                            j, k, lookup_table
                        )
                        index_ik = combinatorial_number_system_d1_forward(
                            i, k, lookup_table
                        )

                        if (
                            cochain[index_ij]
                            + cochain[index_jk]
                            - cochain[index_ik]
                            != 0
                        ):
                            is_cocycle = False
                            first_failure = min(
                                first_failure,
                                dist_mat[i, j],
                                dist_mat[j, k],
                                dist_mat[i, k],
                            )
    return is_cocycle, first_failure


# improve circular coordinates with lattice reduction
def _reduce_circular_coordinates(circ_coords, gram_matrix):
    lattice_red_input = np.linalg.cholesky(gram_matrix)
    decorrelated_vectors, change_basis = _lll(lattice_red_input.T)
    change_basis = change_basis.T
    new_circ_coords = CircleMapUtils.linear_combination(
        np.array(circ_coords), change_basis
    )
    return new_circ_coords, change_basis, decorrelated_vectors


def _gram_matrix(vectors, inner_product):
    n = len(vectors)
    res = np.zeros((n, n))
    for i, v in enumerate(vectors):
        for j, w in enumerate(vectors):
            res[i, j] = v.T @ inner_product @ w
    return res


## Lattice Reduction


# Gram-Schmidt (without normalization)
def _gram_schmidt(B):
    def gs_cofficient(v1, v2):
        return np.dot(v2, v1) / np.dot(v1, v1)

    # projects v2 onto v1
    def proj(v1, v2):
        return gs_cofficient(v1, v2) * v1

    n = len(B)
    A = np.zeros((n, n))
    A[:, 0] = B[:, 0]
    for i in range(1, n):
        Ai = B[:, i]
        for j in range(0, i):
            Aj = B[:, j]
            # t = np.dot(B[i],B[j])
            Ai = Ai - proj(Aj, Ai)
        A[:, i] = Ai
    return A


# LLL algorithm
def _lll(B, delta=3 / 4):
    B = B.copy()
    Q = _gram_schmidt(B)
    change = np.eye(B.shape[0])

    def mu(i, j):
        v = B[:, i]
        u = Q[:, j]
        return np.dot(v, u) / np.dot(u, u)

    n, k = len(B), 1
    while k < n:
        # length reduction step
        for j in reversed(range(k)):
            if abs(mu(k, j)) > 0.5:
                mu_kj = mu(k, j)
                B[:, k] = B[:, k] - round(mu_kj) * B[:, j]
                change[:, k] = change[:, k] - round(mu_kj) * change[:, j]
                Q = _gram_schmidt(B)

        # swap step
        if np.dot(Q[:, k], Q[:, k]) > (delta - mu(k, k - 1) ** 2) * (
            np.dot(Q[:, k - 1], Q[:, k - 1])
        ):
            k = k + 1
        else:
            B_k = B[:, k].copy()
            B_k1 = B[:, k - 1].copy()
            B[:, k], B[:, k - 1] = B_k1, B_k
            change_k = change[:, k].copy()
            change_k1 = change[:, k - 1].copy()
            change[:, k], change[:, k - 1] = change_k1, change_k

            Q = _gram_schmidt(B)
            k = max(k - 1, 1)

    return B, change



import warnings
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.offsetbox import OffsetImage, AnnotationBbox


class PlotUtils:
    """
    Plotting utilities for DREiMac's output.

    """

    @staticmethod
    def imscatter(X, P, dim, zoom=1, ax=None):
        """
        Plot patches in specified locations in R2

        Parameters
        ----------
        X : ndarray (N, 2)
            The positions of each patch in R2
        P : ndarray (N, dim*dim)
            An array of all of the patches
        dim : int
            The dimension of each patch
        ax : matplotlib axes, optional
            If given, plot on those axes, otherwise plot
            on current axes by calling gca()

        """
        # https://stackoverflow.com/questions/22566284/matplotlib-how-to-plot-images-instead-of-points

        ax = ax or plt.gca()
        min_color = np.min(P)
        max_color = np.max(P)
        for i in range(P.shape[0]):
            patch = np.reshape(P[i, :], (dim, dim))
            x, y = X[i, :]
            im = OffsetImage(patch, zoom=zoom, cmap="gray")
            im.get_children()[0].set_clim(vmin=min_color, vmax=max_color)
            ab = AnnotationBbox(im, (x, y), xycoords="data", frameon=False)
            ax.add_artist(ab)
        ax.update_datalim(X)
        ax.autoscale()
        ax.set_xticks([])
        ax.set_yticks([])
        return ax

    @staticmethod
    def plot_patches(P, zoom=1, ax=None):
        """
        Plot patches in a best fitting rectangular grid

        Parameters
        ----------
        ax : matplotlib axes, optional
            If given, plot on those axes, otherwise plot
            on current axes by calling gca()

        """
        N = P.shape[0]
        d = int(np.sqrt(P.shape[1]))
        dgrid = int(np.ceil(np.sqrt(N)))
        ex = np.arange(dgrid)
        x, y = np.meshgrid(ex, ex)
        X = np.zeros((N, 2))
        X[:, 0] = x.flatten()[0:N]
        X[:, 1] = y.flatten()[0:N]
        return PlotUtils.imscatter(X, P, d, zoom, ax)

    @staticmethod
    def plot_proj_boundary(ax=None):
        """
        Plot the boundary of RP2 on the unit circle

        Parameters
        ----------
        ax : matplotlib axes, optional
            If given, plot on those axes, otherwise plot
            on current axes by calling gca()

        """

        ax = ax or plt.gca()
        t = np.linspace(0, 2 * np.pi, 200)
        ax.plot(np.cos(t), np.sin(t), "c")
        ax.axis("equal")
        ax.arrow(
            -0.1, 1, 0.001, 0, head_width=0.15, head_length=0.2, fc="c", ec="c", width=0
        )
        ax.arrow(
            0.1,
            -1,
            -0.001,
            0,
            head_width=0.15,
            head_length=0.2,
            fc="c",
            ec="c",
            width=0,
        )
        ax.set_facecolor((0.35, 0.35, 0.35))
        return ax

    @staticmethod
    def plot_2sphere_boundary(ax=None):
        """
        Plot the boundary of two sphere hemispheres

        Parameters
        ----------
        ax : matplotlib axes, optional
            If given, plot on those axes, otherwise plot
            on current axes by calling gca()

        """

        second_circle_offset = 2.5
        ax = ax or plt.gca()
        t = np.linspace(0, 2 * np.pi, 200)
        ax.plot(np.cos(t), np.sin(t), "c")
        ax.plot(np.cos(t) + second_circle_offset, np.sin(t), "c")
        ax.axis("equal")

        ax.arrow(
            0, 0.9, 0, 0.001, head_width=0.15, head_length=0.2, fc="c", ec="c", width=0
        )
        ax.arrow(
            second_circle_offset,
            1.1,
            0,
            -0.001,
            head_width=0.15,
            head_length=0.2,
            fc="c",
            ec="c",
            width=0,
        )

        ax.arrow(
            0,
            -0.9,
            0,
            -0.001,
            head_width=0.15,
            head_length=0.2,
            fc="c",
            ec="c",
            width=0,
        )
        ax.arrow(
            second_circle_offset,
            -1.1,
            0,
            0.001,
            head_width=0.15,
            head_length=0.2,
            fc="c",
            ec="c",
            width=0,
        )

        ax.arrow(
            0.9, 0, 0.001, 0, head_width=0.15, head_length=0.2, fc="c", ec="c", width=0
        )
        ax.arrow(
            second_circle_offset + 1.1,
            0,
            -0.001,
            0,
            head_width=0.15,
            head_length=0.2,
            fc="c",
            ec="c",
            width=0,
        )

        ax.arrow(
            -0.9,
            0,
            -0.001,
            0,
            head_width=0.15,
            head_length=0.2,
            fc="c",
            ec="c",
            width=0,
        )
        ax.arrow(
            second_circle_offset - 1.1,
            0,
            0.001,
            0,
            head_width=0.15,
            head_length=0.2,
            fc="c",
            ec="c",
            width=0,
        )

        ax.set_facecolor((0.35, 0.35, 0.35))

        return ax

    @staticmethod
    def plot_3sphere_mesh(n_parallels = 10, n_meridians = 20, alpha=0.1, ax=None):
        """
        Draw a spherical mesh in 3D around the unit sphere.

        Parameters
        ----------
        n_parallels : int
            How many parallels to draw.

        n_meridians : int
            How many meridians to draw.

        alpha : float
            Opacity of the mesh. Must be between 0 and 1.

        ax : matplotlib axes, optional
            If given, plot on those axes, otherwise plot
            on current axes by calling gca()

        """

        ax = ax or plt.gca()
        u, v = np.mgrid[0 : 2 * np.pi : n_meridians * 1j, 0 : np.pi : n_parallels * 1j]
        x = np.cos(u) * np.sin(v)
        y = np.sin(u) * np.sin(v)
        z = np.cos(v)
        ax.plot_wireframe(x, y, z, color="black", alpha=alpha)
        return ax

    @staticmethod
    def set_axes_equal(ax):
        """
        Make axes of 3D plot have equal scale so that spheres appear as spheres,
        cubes as cubes, etc..  This is one possible solution to Matplotlib's
        ax.set_aspect('equal') and ax.axis('equal') not working for 3D.

        Parameters
        ----------
        ax : a matplotlib axis, e.g., as output from plt.gca().

        """
        # taken from https://stackoverflow.com/a/31364297/2171328

        x_limits = ax.get_xlim3d()
        y_limits = ax.get_ylim3d()
        z_limits = ax.get_zlim3d()

        x_range = abs(x_limits[1] - x_limits[0])
        x_middle = np.mean(x_limits)
        y_range = abs(y_limits[1] - y_limits[0])
        y_middle = np.mean(y_limits)
        z_range = abs(z_limits[1] - z_limits[0])
        z_middle = np.mean(z_limits)

        # The plot bounding box is a sphere in the sense of the infinity
        # norm, hence I call half the max range the plot radius.
        plot_radius = 0.5 * max([x_range, y_range, z_range])

        ax.set_xlim3d([x_middle - plot_radius, x_middle + plot_radius])
        ax.set_ylim3d([y_middle - plot_radius, y_middle + plot_radius])
        ax.set_zlim3d([z_middle - plot_radius, z_middle + plot_radius])
        return ax

    @staticmethod
    def plot_rp2_stereo(S, f, arrowcolor="c", facecolor=(0.15, 0.15, 0.15)):
        """
        Plot a 2D Stereographic Projection

        Parameters
        ----------
        S : ndarray (N, 2)
            An Nx2 array of N points to plot on RP2
        f : ndarray (N) or ndarray (N, 3)
            A function with which to color the points, or a list of colors

        """

        def _plot_rp2_circle(
            ax, arrowcolor="c", facecolor=(0.15, 0.15, 0.15), do_arrows=True, pad=1.1
        ):
            """
            Plot a circle with arrows showing the identifications for RP2.
            Set an equal aspect ratio and get rid of the axis ticks, since
            they are clear from the circle

            Parameters
            ----------
            ax: matplotlib axis
                Axis onto which to plot the circle+arrows
            arrowcolor: string or ndarray(3) or ndarray(4)
                Color for the circle and arrows
            facecolor: string or ndarray(3) or ndarray(4)
                Color for background of the plot
            do_arrows: boolean
                Whether to draw the arrows
            pad: float
                The dimensions of the window around the unit square

            """
            t = np.linspace(0, 2 * np.pi, 200)
            ax.plot(np.cos(t), np.sin(t), c=arrowcolor)
            ax.axis("equal")
            ax = plt.gca()
            if do_arrows:
                ax.arrow(
                    -0.1,
                    1,
                    0.001,
                    0,
                    head_width=0.15,
                    head_length=0.2,
                    fc=arrowcolor,
                    ec=arrowcolor,
                    width=0,
                )
                ax.arrow(
                    0.1,
                    -1,
                    -0.001,
                    0,
                    head_width=0.15,
                    head_length=0.2,
                    fc=arrowcolor,
                    ec=arrowcolor,
                    width=0,
                )
            ax.set_facecolor(facecolor)
            ax.set_xticks([])
            ax.set_yticks([])
            ax.set_xlim([-pad, pad])
            ax.set_ylim([-pad, pad])

        if not (S.shape[1] == 2):
            warnings.warn(
                "Plotting stereographic RP2 projection, but points are not 2 dimensional"
            )
        _plot_rp2_circle(plt.gca(), arrowcolor, facecolor)
        if f.size > S.shape[0]:
            plt.scatter(S[:, 0], S[:, 1], 20, c=f, cmap="afmhot")
        else:
            plt.scatter(S[:, 0], S[:, 1], 20, f, cmap="afmhot")

    @staticmethod
    def plot_rp3_stereo(ax, S, f, draw_sphere=False):
        """
        Plot a 3D Stereographic Projection

        Parameters
        ----------
        ax : matplotlib axis
            3D subplotting axis
        S : ndarray (N, 3)
            An Nx3 array of N points to plot on RP3
        f : ndarray (N) or ndarray (N, 3)
            A function with which to color the points, or a list of colors
        draw_sphere : boolean
            Whether to draw the 2-sphere

        """
        if not (S.shape[1] == 3):
            warnings.warn(
                "Plotting stereographic RP3 projection, but points are not 4 dimensional"
            )
        if f.size > S.shape[0]:
            ax.scatter(S[:, 0], S[:, 1], S[:, 2], c=f, cmap="afmhot")
        else:
            c = plt.get_cmap("afmhot")
            C = f - np.min(f)
            C = C / np.max(C)
            C = c(np.array(np.round(C * 255), dtype=np.int32))
            C = C[:, 0:3]
            ax.scatter(S[:, 0], S[:, 1], S[:, 2], c=C, cmap="afmhot")
        if draw_sphere:
            u, v = np.mgrid[0 : 2 * np.pi : 20j, 0 : np.pi : 20j]
            ax.set_aspect("equal")
            x = np.cos(u) * np.sin(v)
            y = np.sin(u) * np.sin(v)
            z = np.cos(v)
            ax.plot_wireframe(x, y, z, color="k")



import numpy as np

from dreimac import ProjectiveCoords, ComplexProjectiveCoords, GeometryExamples

class TestRealProjective:

    def test_rp2(self):
        """
        Test projective coordinates on the projective plane
        Parameters
        ----------
        NSamples : int
            Number of random samples on the projective plane
        NLandmarks : int
            Number of landmarks to take in the projective coordinates computation
        """
        n_samples = 10000
        n_landmarks = 100
        X, D = GeometryExamples.rp2_metric(n_samples)

        ## Coming up with ground truth theta and phi for RP2 for colors
        #SOrig = ProjectiveCoords.get_stereo_proj_codim1(X)
        #phi = np.sqrt(np.sum(SOrig**2, 1))
        #theta = np.arccos(np.abs(SOrig[:, 0]))
        
        pc = ProjectiveCoords(D, n_landmarks, distance_matrix=True, verbose=True)
        coordinates = pc.get_coordinates()
        assert len(coordinates) == len(X)
        total_variance = np.linalg.norm(pc.variance_)
        target_coordinates = 2
        variance_threshold = 0.8
        assert np.linalg.norm(pc.variance_[:target_coordinates]) >= total_variance * variance_threshold



    def test_klein_bottle(self):
        """
        Test projective coordinates on the Klein bottle
        """
        n_samples = 10000
        n_landmarks = 200
        R = 2
        r = 1
        X = GeometryExamples.klein_bottle_4d(n_samples, R, r)
        pc = ProjectiveCoords(X, n_landmarks, verbose=True)
        coordinates = pc.get_coordinates()
        assert len(coordinates) == len(X)
        total_variance = np.linalg.norm(pc.variance_)
        target_coordinates = 2
        variance_threshold = 0.3
        assert np.linalg.norm(pc.variance_[:target_coordinates]) >= total_variance * variance_threshold


    def test_line_segment(self):
        """
        Test projective coordinates on a set of line segment patches
        """
        dim = 10
        P = GeometryExamples.line_patches(dim=dim, n_angles=200, n_offsets=200, sigma=0.25)
        pc = ProjectiveCoords(P, n_landmarks=100)
        coordinates = pc.get_coordinates()
        assert len(coordinates) == len(P)
        total_variance = np.linalg.norm(pc.variance_)
        target_coordinates = 2
        variance_threshold = 0.3
        assert np.linalg.norm(pc.variance_[:target_coordinates]) >= total_variance * variance_threshold


    def test_image_patches(self):
        X = GeometryExamples.line_patches(dim=10, n_angles=200, n_offsets=200, sigma=0.25)
        proj = ProjectiveCoords(X, n_landmarks=200)
        coordinates = proj.get_coordinates(proj_dim=2, perc=0.8, cocycle_idx=0)
        assert len(coordinates) == len(X)
        total_variance = np.linalg.norm(proj.variance_)
        target_coordinates = 2
        variance_threshold = 0.3
        assert np.linalg.norm(proj.variance_[:target_coordinates]) >= total_variance * variance_threshold


class TestComplexProjective:

    def test_sphere(self):
        data = GeometryExamples.sphere(2000)
        cpc = ComplexProjectiveCoords(data, n_landmarks=100)
        target_coordinates = 1
        coordinates = cpc.get_coordinates(proj_dim=target_coordinates)
        assert len(coordinates) == len(data)
        total_variance = np.linalg.norm(cpc.variance_)
        variance_threshold = 0.3
        assert np.linalg.norm(cpc.variance_[:target_coordinates]) >= total_variance * variance_threshold

    def test_moving_dot(self):
        P = GeometryExamples.moving_dot(40)
        cpc = ComplexProjectiveCoords(P, n_landmarks=300)
        target_coordinates = 1
        coordinates = cpc.get_coordinates(proj_dim=target_coordinates)
        assert len(coordinates) == len(P)
        total_variance = np.linalg.norm(cpc.variance_)
        variance_threshold = 0.3
        assert np.linalg.norm(cpc.variance_[:target_coordinates]) >= total_variance * variance_threshold



import numpy as np
from dreimac.combinatorial import (
    combinatorial_number_system_table,
    combinatorial_number_system_forward,
    combinatorial_number_system_d2_forward,
)

class TestCombinatorialNumberSystem:

    def test_forward(self):

        lookup_table = combinatorial_number_system_table(5, 2)

        check = [
            [[0, 1, 2], 0],
            [[0, 1, 3], 1],
            [[0, 2, 3], 2],
            [[1, 2, 3], 3],
            [[0, 1, 4], 4],
        ]

        for simplex, answer in check:
            assert answer == combinatorial_number_system_forward(
                np.array(simplex), lookup_table
            )
            assert combinatorial_number_system_d2_forward(
                simplex[0], simplex[1], simplex[2], lookup_table
            ) == combinatorial_number_system_forward(np.array(simplex), lookup_table)



from .circularcoords import CircularCoords
from .toroidalcoords import ToroidalCoords
from .projectivecoords import ProjectiveCoords
from .complexprojectivecoords import ComplexProjectiveCoords
from .lenscoords import LensCoords
from .utils import GeometryUtils, PartUnity, GeometryExamples, CircleMapUtils, ProjectiveMapUtils, LensMapUtils
from .plotting_utils import PlotUtils

from ._version import __version__



import numpy as np
from numba import jit
import scipy.sparse as sparse
from scipy.sparse.linalg import lsqr
from scipy.optimize import LinearConstraint, milp
from .utils import PartUnity, CohomologyUtils, EquivariantPCA
from .emcoords import EMCoords
from .combinatorial import (
    combinatorial_number_system_table,
    combinatorial_number_system_d1_forward,
    combinatorial_number_system_d2_forward,
)


class ComplexProjectiveCoords(EMCoords):
    """
    Object that performs multiscale complex projective coordinates via
    persistent cohomology of sparse filtrations (Perea 2018).


    Parameters
    ----------
    X: ndarray(N, d)
        A point cloud with N points in d dimensions
    n_landmarks: int
        Number of landmarks to use
    distance_matrix: boolean
        If true, treat X as a distance matrix instead of a point cloud
    prime : int
        Field coefficient with which to compute rips on landmarks
    maxdim : int
        Maximum dimension of homology.  Only dimension 2 is needed for complex projective coordinates,
        but it may be of interest to see other dimensions
    """

    def __init__(
        self, X, n_landmarks, distance_matrix=False, prime=41, maxdim=2, verbose=False
    ):
        EMCoords.__init__(self, X, n_landmarks, distance_matrix, prime, maxdim, verbose)
        simplicial_complex_dimension = 3
        self.cns_lookup_table_ = combinatorial_number_system_table(
            n_landmarks, simplicial_complex_dimension
        )
        self.type_ = "complexprojective"

    def get_coordinates(
        self,
        perc=0.5,
        cocycle_idx=0,
        proj_dim=1,
        partunity_fn=PartUnity.linear,
        standard_range=True,
        check_cocycle_condition=True,
        projective_dim_red_mode="exponential",
    ):
        """
        Get complex projective coordinates.


        Parameters
        ----------
        perc : float
            Percent coverage. Must be between 0 and 1.
        cocycle_idx : integer
            Integer representing the index of the persistent cohomology class
            used to construct the Eilenberg-MacLane coordinate. Persistent cohomology
            classes are ordered by persistence, from largest to smallest.
        proj_dim : integer
            Complex dimension down to which to project the data.
        partunity_fn : (dist_land_data, r_cover) -> phi
            A function from the distances of each landmark to a bump function
        standard_range : bool
            Whether to use the parameter perc to choose a filtration parameter that guarantees
            that the selected cohomology class represents a non-trivial class in the Cech complex.
        check_cocycle_condition : bool
            Whether to check, and fix if necessary, that the integer cocycle constructed
            using finite field coefficients satisfies the cocycle condition.
        projective_dim_red_mode : string
            Either "one-by-one", "exponential", or "direct". How to perform equivariant
            dimensionality reduction. "exponential" usually works best, being fast
            without compromising quality.

        Returns
        -------
        ndarray(N, proj_dim+1)
            Complex projective coordinates
        """

        homological_dimension = 2
        cohomdeath_rips, cohombirth_rips, cocycle = self.get_representative_cocycle(
            cocycle_idx, homological_dimension
        )

        # determine radius for balls
        r_cover, rips_threshold = EMCoords.get_cover_radius(
            self, perc, cohomdeath_rips, cohombirth_rips, standard_range
        )

        # compute partition of unity and choose a cover element for each data point
        varphi, ball_indx = EMCoords.get_covering_partition(self, r_cover, partunity_fn)

        # compute boundary matrix
        delta1 = CohomologyUtils.make_delta1(
            self.dist_land_land_, rips_threshold, self.cns_lookup_table_
        )

        # lift to integer cocycles
        integer_cocycle = CohomologyUtils.lift_to_integer_cocycle(
            cocycle, prime=self.prime_
        )

        # go from sparse to dense representation of cocycles
        integer_cocycle_as_vector = CohomologyUtils.sparse_cocycle_to_vector(
            integer_cocycle, self.cns_lookup_table_, self.n_landmarks_, int
        )

        if check_cocycle_condition:
            is_cocycle = _is_two_cocycle(
                integer_cocycle_as_vector,
                self.dist_land_land_,
                rips_threshold,
                self.cns_lookup_table_,
            )
            if not is_cocycle:
                delta2 = CohomologyUtils.make_delta2_compact(
                    self.dist_land_land_, rips_threshold, self.cns_lookup_table_
                )
                d2cocycle = delta2 @ integer_cocycle_as_vector.T

                y = d2cocycle // self.prime_

                constraints = LinearConstraint(delta2, y, y, keep_feasible=True)
                n_edges = delta2.shape[1]
                objective = np.zeros((n_edges), dtype=int)
                integrality = np.ones((n_edges), dtype=int)
                optimizer_solution = milp(
                    objective,
                    integrality=integrality,
                    constraints=constraints,
                )

                if not optimizer_solution["success"]:
                    raise Exception(
                        "The cohomology class at index "
                        + str(cocycle_idx)
                        + " does not have an integral lift."
                    )
                else:
                    solution = optimizer_solution["x"]
                    new_cocycle_as_vector = (
                        integer_cocycle_as_vector
                        - self.prime_ * np.array(np.rint(solution), dtype=int)
                    )
                    integer_cocycle_as_vector = new_cocycle_as_vector

        # compute harmonic representatives of cocycles and their projective-valued integrals
        integral = lsqr(delta1, integer_cocycle_as_vector)[0]
        harmonic_representative = integer_cocycle_as_vector - delta1 @ integral

        # compute complex projective coordinates on data points
        class_map = _sparse_integrate(
            harmonic_representative,
            integral,
            varphi,
            ball_indx,
            self.dist_land_land_,
            rips_threshold,
            self.cns_lookup_table_,
        )

        # reduce dimensionality of complex projective space
        epca = EquivariantPCA.ppca(
            class_map,
            proj_dim,
            projective_dim_red_mode,
            self.verbose,
        )
        self.variance_ = epca["variance"]

        return epca["X"]


# turn cocycle into tensor
def _two_cocycle_to_tensor(
    cocycle: np.ndarray,
    dist_mat: np.ndarray,
    threshold: float,
    lookup_table: np.ndarray,
):
    n_points = dist_mat.shape[0]

    res = np.zeros((n_points, n_points, n_points))

    @jit(fastmath=True, nopython=True)
    def _get_res(
        cocycle: np.ndarray,
        dist_mat: np.ndarray,
        threshold: float,
        lookup_table: np.ndarray,
        n_points: int,
        res: np.ndarray,
    ):
        for i in range(n_points):
            for j in range(i + 1, n_points):
                if dist_mat[i, j] < threshold:
                    for k in range(j + 1, n_points):
                        if dist_mat[i, k] < threshold and dist_mat[j, k] < threshold:
                            flat_index = combinatorial_number_system_d2_forward(
                                i, j, k, lookup_table
                            )
                            val = cocycle[flat_index]
                            # 012
                            res[i, j, k] = val
                            # 021
                            res[i, k, j] = -val
                            # 102
                            res[j, i, k] = -val
                            # 210
                            res[k, j, i] = -val
                            # 201
                            res[k, i, j] = val
                            # 120
                            res[j, k, i] = val

    _get_res(cocycle, dist_mat, threshold, lookup_table, n_points, res)

    return res


def _one_cocycle_to_tensor(
    cocycle: np.ndarray,
    dist_mat: np.ndarray,
    threshold: float,
    lookup_table: np.ndarray,
):
    n_points = dist_mat.shape[0]

    res = np.zeros((n_points, n_points))

    @jit(fastmath=True, nopython=True)
    def _get_res(
        cocycle: np.ndarray,
        dist_mat: np.ndarray,
        threshold: float,
        lookup_table: np.ndarray,
        n_points: int,
        res: np.ndarray,
    ):
        for i in range(n_points):
            for j in range(i + 1, n_points):
                if dist_mat[i, j] < threshold:
                    flat_index = combinatorial_number_system_d1_forward(
                        i, j, lookup_table
                    )
                    val = cocycle[flat_index]
                    res[i, j] = val
                    res[j, i] = -val

    _get_res(cocycle, dist_mat, threshold, lookup_table, n_points, res)

    return res


def _sparse_integrate(
    harm_rep,
    integral,
    part_unity,
    membership_function,
    dist_mat,
    threshold,
    lookup_table,
):
    nu = _one_cocycle_to_tensor(integral, dist_mat, threshold, lookup_table)

    eta = _two_cocycle_to_tensor(
        harm_rep,
        dist_mat,
        threshold,
        lookup_table,
    )

    class_map0 = np.zeros_like(part_unity.T)

    @jit(nopython=True)
    def _assemble(
        class_map: np.ndarray,
        nu: np.ndarray,
        eta: np.ndarray,
        varphi: np.ndarray,
        n_landmarks: int,
        n_data: int,
        ball_indx: np.ndarray,
    ):
        for b in range(n_data):
            for i in range(n_landmarks):
                class_map[b, i] += nu[i, ball_indx[b]]
                for t in range(n_landmarks):
                    class_map[b, i] += varphi[t, b] * eta[i, ball_indx[b], t]
        return np.exp(2 * np.pi * 1j * class_map0) * np.sqrt(varphi.T)

    return _assemble(
        class_map0,
        nu,
        eta,
        part_unity,
        dist_mat.shape[0],
        part_unity.shape[1],
        membership_function,
    )


@jit(fastmath=True, nopython=True)
def _is_two_cocycle(
    cochain: np.ndarray,
    dist_mat: np.ndarray,
    threshold: float,
    lookup_table: np.ndarray,
):
    is_cocycle = True
    n_points = dist_mat.shape[0]
    for i in range(n_points):
        for j in range(i + 1, n_points):
            if dist_mat[i, j] < threshold:
                for k in range(j + 1, n_points):
                    if dist_mat[i, k] < threshold and dist_mat[j, k] < threshold:
                        for l in range(k + 1, n_points):
                            if (
                                dist_mat[i, l] < threshold
                                and dist_mat[j, l] < threshold
                                and dist_mat[k, l] < threshold
                            ):
                                index_ijk = combinatorial_number_system_d2_forward(
                                    i, j, k, lookup_table
                                )
                                index_ijl = combinatorial_number_system_d2_forward(
                                    i, j, l, lookup_table
                                )
                                index_ikl = combinatorial_number_system_d2_forward(
                                    i, k, l, lookup_table
                                )
                                index_jkl = combinatorial_number_system_d2_forward(
                                    j, k, l, lookup_table
                                )

                                if (
                                    cochain[index_ijk]
                                    - cochain[index_ijl]
                                    + cochain[index_ikl]
                                    - cochain[index_jkl]
                                    != 0
                                ):
                                    is_cocycle = False
    return is_cocycle



from .utils import PartUnity
from .toroidalcoords import ToroidalCoords


class CircularCoords(ToroidalCoords):
    """
    Object that performs circular coordinates via persistent cohomology of
    sparse filtrations (Jose Perea 2020).

    Parameters
    ----------
    X: ndarray(N, d)
        A point cloud with N points in d dimensions
    n_landmarks: int
        Number of landmarks to use
    distance_matrix: boolean
        If true, treat X as a distance matrix instead of a point cloud
    prime : int
        Field coefficient with which to compute rips on landmarks
    maxdim : int
        Maximum dimension of homology.  Only dimension 1 is needed for circular coordinates,
        but it may be of interest to see other dimensions (e.g. for a torus)
    verbose : bool
        Print debug information.
    """

    def __init__(
        self, X, n_landmarks, distance_matrix=False, prime=41, maxdim=1, verbose=False
    ):
        ToroidalCoords.__init__(
            self,
            X=X,
            n_landmarks=n_landmarks,
            distance_matrix=distance_matrix,
            prime=prime,
            maxdim=maxdim,
            verbose=verbose,
        )
        self.type_ = "circ"

    def get_coordinates(
        self,
        perc=0.5,
        cocycle_idx=0,
        partunity_fn=PartUnity.linear,
        standard_range=True,
        check_cocycle_condition=True,
    ):
        """
        Get circular coordinates.


        Parameters
        ----------
        perc : float
            Percent coverage. Must be between 0 and 1.
        cocycle_idx : integer
            Integer representing the index of the persistent cohomology class
            used to construct the Eilenberg-MacLane coordinate. Persistent cohomology
            classes are ordered by persistence, from largest to smallest.
        partunity_fn : (dist_land_data, r_cover) -> phi
            A function from the distances of each landmark to a bump function
        standard_range : bool
            Whether to use the parameter perc to choose a filtration parameter that guarantees
            that the selected cohomology class represents a non-trivial class in the Cech complex.
        check_cocycle_condition : bool
            Whether to check, and fix if necessary, that the integer cocycle constructed
            using finite field coefficients satisfies the cocycle condition.

        Returns
        -------
        thetas : ndarray(N)
            Circular coordinates
        """

        return ToroidalCoords.get_coordinates(
            self,
            perc,
            [cocycle_idx],
            partunity_fn,
            standard_range,
            check_cocycle_condition,
        )[0]



import numpy as np
from scipy.spatial import KDTree

from dreimac import CircularCoords, ToroidalCoords, GeometryExamples


def _less_than_or_equal_with_tolerance(x, y):
    return np.allclose(x, y) or x <= y


class TestCircular:
    def test_toroidal_coordinates_independent(self):
        """
        Test that toroidal coordinates returns two circle-valued maps, each one constant in
        one of the two circles.
        """
        prime = 41
        np.random.seed(2)
        N = 500
        X = np.zeros((N * 2, 2))
        t = np.linspace(0, 1, N + 1)[0:N] ** 1.2
        t = 2 * np.pi * t
        X[0:N, 0] = np.cos(t)
        X[0:N, 1] = np.sin(t)
        X[N::, 0] = 2 * np.cos(t) + 4
        X[N::, 1] = 2 * np.sin(t) + 4
        X = X + 0.2 * np.random.randn(X.shape[0], 2)

        tc = ToroidalCoords(X, 200, prime=prime)
        coords1, coords2 = tc.get_coordinates(cocycle_idxs=[0, 1])

        assert len(coords1) == len(X)
        assert len(coords2) == len(X)
        print(
            _maximum_circle_distance(coords1[:N]),
            _maximum_circle_distance(coords2[N:]),
            _maximum_circle_distance(coords2[:N]),
            _maximum_circle_distance(coords1[N:]),
        )
        assert (
            np.isclose(_maximum_circle_distance(coords1[:N]), 0)
            and np.isclose(_maximum_circle_distance(coords2[N:]), 0)
        ) or (
            np.isclose(_maximum_circle_distance(coords2[:N]), 0)
            and np.isclose(_maximum_circle_distance(coords1[N:]), 0)
        )

    def test_toroidal_coordinates_less_energy(self):
        """
        Test that toroidal coordinates returns circle-valued maps with less Dirichlet
        energy in total.
        """
        # TODO: use the following instead
        # try:
        #   ...
        # except ...:
        #   self.fail(...)
        X = GeometryExamples.genus_two_surface()
        tc = ToroidalCoords(X, n_landmarks=1000)
        cocycle_idxs = [0, 1, 2, 3]
        toroidal_coords = tc.get_coordinates(cocycle_idxs=cocycle_idxs)

        assert toroidal_coords.shape == (4, X.shape[0])
        assert _less_than_or_equal_with_tolerance(
            np.linalg.norm(tc.gram_matrix_), np.linalg.norm(tc.original_gram_matrix_)
        )

        X = GeometryExamples.torus_3d(2000, 5, 1, seed=0)
        tc = ToroidalCoords(X, n_landmarks=500)
        cocycle_idxs = [0, 1]
        toroidal_coords = tc.get_coordinates(cocycle_idxs=cocycle_idxs)

        assert toroidal_coords.shape == (2, X.shape[0])
        assert _less_than_or_equal_with_tolerance(
            np.linalg.norm(tc.gram_matrix_), np.linalg.norm(tc.original_gram_matrix_)
        )


        X = GeometryExamples.bullseye()
        tc = ToroidalCoords(X, n_landmarks=300)
        cocycle_idxs = [0, 1, 2]
        toroidal_coords = tc.get_coordinates(cocycle_idxs=cocycle_idxs)

        assert toroidal_coords.shape == (3, X.shape[0])
        assert _less_than_or_equal_with_tolerance(
            np.linalg.norm(tc.gram_matrix_), np.linalg.norm(tc.original_gram_matrix_)
        )


        X = GeometryExamples.three_circles()
        tc = ToroidalCoords(X, n_landmarks=300)
        cocycle_idxs = [0, 1, 2]
        toroidal_coords = tc.get_coordinates(cocycle_idxs=cocycle_idxs)

        assert toroidal_coords.shape == (3, X.shape[0])
        assert _less_than_or_equal_with_tolerance(
            np.linalg.norm(tc.gram_matrix_), np.linalg.norm(tc.original_gram_matrix_)
        )


    def test_trefoil(self):
        """Check that circular coordinates returns a continuous map, even when the lifted
        cochain may fail to be a cocycle and fix the cocycle using the integer linear system
        """
        for noisy in [True, False]:
            X = GeometryExamples.trefoil(n_samples=2500, horizontal_width=10, noisy=noisy)

            prime = 3
            large_perc = 0.8
            cc = CircularCoords(X, 300, prime=prime)
            coords = cc.get_coordinates(
                perc=large_perc,
                cocycle_idx=0,
                check_cocycle_condition=True,
            )
            assert len(coords) == len(X)

            tree = KDTree(X)
            k = 5
            _, nns = tree.query(X, k=k)

            tolerance = 5 / 100 * (2 * np.pi)  # 5% of the full circle

            for i in range(X.shape[0]):
                assert _maximum_circle_distance(coords[nns[i]]) <= tolerance


def _circle_distance(x, y):
    return np.minimum(
        np.minimum(np.abs(x - y), np.abs((x - 2 * np.pi) - y)),
        np.abs(x - (y - 2 * np.pi)),
    )


def _maximum_circle_distance(xs):
    return max([_circle_distance(a, b) for a in xs for b in xs])



"""
A superclass for shared code across all different types of coordinates
"""
import numpy as np
from scipy.sparse.linalg import lsqr
import time
from .utils import CohomologyUtils
from ripser import ripser
import warnings


class EMCoords(object):
    def __init__(self, X, n_landmarks, distance_matrix, prime, maxdim, verbose):
        """
        Perform persistent homology on the landmarks, store distance
        from the landmarks to themselves and to the rest of the points,
        and sort persistence diagrams and cocycles in decreasing order of persistence

        Parameters
        ----------
        X: ndarray(N, d)
            A point cloud with N points in d dimensions, or a matrix of distances from N points to d points.
            See distance_matrix, below, for a description of the second scenario.
        n_landmarks: int
            Number of landmarks to use
        distance_matrix: boolean
            If true, treat X as a distance matrix instead of a point cloud.
            If X is square, then the i-th row should represent the same point as the i-th column, meaning
            that the matrix should be symmetric.
            If X is not square, then it should have more columns than rows (i.e., N > d).
            Moreover, if i < N, the i-th row should represent the same point as the i-th column.
            When X is not square, the rows of X are interpreted as a subsample and the columns as all available points; thus
            X represents the distance from the points in the subsample to all available points.
        prime : int
            Field coefficient with which to compute rips on landmarks
        maxdim : int
            Maximum dimension of homology. 

        """
        assert maxdim >= 1
        self.verbose = verbose
        self.X_ = X
        if verbose:
            tic = time.time()
            print("Doing TDA...")
        if distance_matrix is False:
            ripser_metric_input = X 
        elif X.shape[0] == X.shape[1]:
            ripser_metric_input = X 
        else:
            ripser_metric_input = X[:,:X.shape[0]]
        res = ripser(
            ripser_metric_input,
            distance_matrix=distance_matrix,
            coeff=prime,
            maxdim=maxdim,
            n_perm=n_landmarks,
            do_cocycles=True,
        )
        if verbose:
            print("Elapsed time persistence: %.3g seconds" % (time.time() - tic))
        self.prime_ = prime
        self.dgms_ = res["dgms"]
        self.idx_land_ = res["idx_perm"]
        self.n_landmarks_ = len(self.idx_land_)
        #self.dist_land_data_ = res["dperm2all"]
        if distance_matrix is False:
            self.dist_land_data_ = res["dperm2all"]
        elif X.shape[0] == X.shape[1]:
            self.dist_land_data_ = res["dperm2all"]
        else:
            self.dist_land_data_ = X[self.idx_land_,:]
        self.coverage_ = np.max(np.min(self.dist_land_data_, 1))
        self.dist_land_land_ = self.dist_land_data_[:, self.idx_land_]
        self.cocycles_ = res["cocycles"]
        # Sort persistence diagrams in descending order of persistence
        for i in range(1, maxdim+1):
            idxs = np.argsort(self.dgms_[i][:, 0] - self.dgms_[i][:, 1])
            self.dgms_[i] = self.dgms_[i][idxs, :]
            dgm_lifetime = np.array(self.dgms_[i])
            dgm_lifetime[:, 1] -= dgm_lifetime[:, 0]
            self.cocycles_[i] = [self.cocycles_[i][idx] for idx in idxs]
        CohomologyUtils.reindex_cocycles(self.cocycles_, self.idx_land_, X.shape[0])

        self.type_ = "emcoords"

    def get_representative_cocycle(self, cohomology_class, homological_dimension):
        """
        Compute the representative cocycle, given a list of cohomology classes

        Parameters
        ----------
        cohomology_class : integer
            Integer representing the index of the persistent cohomology class.
            Persistent cohomology classes are ordered by persistence, from largest to smallest.

        Returns
        -------
        cohomdeath: float
            Cohomological death of the linear combination or single cocycle
        cohombirth: float
            Cohomological birth of the linear combination or single cocycle
        cocycle: ndarray(K, homological_dimension+2, dtype=int)
            Representative cocycle. First homological_dimension+1 columns are vertex indices,
            and last column takes values in finite field corresponding to self.prime_
        """

        assert isinstance(cohomology_class, int)

        dgm = self.dgms_[homological_dimension]
        cocycles = self.cocycles_[homological_dimension]
        return (
            dgm[cohomology_class, 0],
            dgm[cohomology_class, 1],
            cocycles[cohomology_class],
        )

    def get_cover_radius(self, perc, cohomdeath_rips, cohombirth_rips, standard_range):
        """
        Determine radius for covering balls

        Parameters
        ----------
        perc : float
            Percent coverage
        cohomdeath: float
            Cohomological death
        cohombirth: float
            Cohomological birth
        standard_range: float
            Whether or not to use the range that guarantees that the cohomology class selected
            is a non-trivial cohomology class in the Cech complex. If False, the class is only
            guaranteed to be non-trivial in the Rips complex.

        Returns
        -------
        r_cover : float
        rips_threshold : float

        """
        start = 2*cohomdeath_rips if standard_range else cohomdeath_rips
        end = cohombirth_rips
        if start > end:
            raise Exception(
                "The cohomology class selected is too short, try setting standard_range to False."
            )
        self.rips_threshold_ = (1 - perc) * start + perc * end
        self.r_cover_ = self.rips_threshold_ / 2

        return self.r_cover_, self.rips_threshold_

    def get_covering_partition(self, r_cover, partunity_fn):
        """
        Create the open covering U = {U_1,..., U_{s+1}} and partition of unity

        Parameters
        ----------
        r_cover: float
            Covering radius
        partunity_fn: (dist_land_data, r_cover) -> phi
            A function from the distances of each landmark to a bump function

        Returns
        -------
        varphi: ndarray(n_data, dtype=float)
            varphi_j(b) = phi_j(b)/(phi_1(b) + ... + phi_{n_landmarks}(b)),
        ball_indx: ndarray(n_data, dtype=int)
            The index of the first open set each data point belongs to
        """
        dist_land_data = self.dist_land_data_
        U = dist_land_data < r_cover
        phi = np.zeros_like(dist_land_data)
        phi[U] = partunity_fn(dist_land_data[U], r_cover)
        # Compute the partition of unity
        # varphi_j(b) = phi_j(b)/(phi_1(b) + ... + phi_{n_landmarks}(b))
        denom = np.sum(phi, 0)
        nzero = np.sum(denom == 0)
        if nzero > 0:
            warnings.warn("There are {} point not covered by a landmark".format(nzero))
            denom[denom == 0] = 1
        varphi = phi / denom[None, :]
        # To each data point, associate the index of the first open set it belongs to
        ball_indx = np.argmax(U, 0)
        return varphi, ball_indx



from setuptools import setup

## Get version information from _version.py
import re

VERSIONFILE = "dreimac/_version.py"
verstrline = open(VERSIONFILE, "rt").read()
VSRE = r"^__version__ = ['\"]([^'\"]*)['\"]"
mo = re.search(VSRE, verstrline, re.M)
if mo:
    verstr = mo.group(1)
else:
    raise RuntimeError("Unable to find version string in %s." % (VERSIONFILE,))

# Use README.md as the package long description
with open("README.md") as f:
    long_description = f.read()


# get requirements
def requirements():
    with open("requirements.txt") as f:
        return [line.strip() for line in f if line.strip()]


setup(
    name="dreimac",
    version=verstr,
    description="DREiMac: Dimensionality reduction with Eilenberg-MacLane coordinates",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Jose A. Perea, Luis Scoccola, Chris Tralie",
    author_email="ctralie@alumni.princeton.edu",
    license="Apache2",
    packages=["dreimac"],
    install_requires=requirements(),
    extras_require={
        "testing": ["pytest"],
        "examples": [],
    },
    python_requires=">=3.8",
    classifiers=[
        "Intended Audience :: Science/Research",
        "Intended Audience :: Education",
        "Intended Audience :: Financial and Insurance Industry",
        "Intended Audience :: Healthcare Industry",
        "Topic :: Scientific/Engineering :: Information Analysis",
        "Topic :: Scientific/Engineering :: Mathematics",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
    ],
    keywords="topological data analysis, dimensionality reduction",
)



# TODO: ideally, these auxiliary methods would be included in utils.py as a CombinatorialNumberSystem class.
# However, currently, numba's support for classes is experimental, so this more robust.
import numpy as np
from scipy.special import comb
from numba import jit


def combinatorial_number_system_table(maximum_number_vertices, maximum_dimension):
    lookup_table = np.zeros((maximum_number_vertices+1, maximum_dimension + 2), dtype=int)

    for i in range(maximum_number_vertices+1):
        for l in range(maximum_dimension + 2):
            lookup_table[i, l] = comb(i, l, exact=True)

    return lookup_table
    
def number_of_simplices_of_dimension(dimension, n_vertices, lookup_table):
    return lookup_table[n_vertices,dimension+1]

@jit(nopython=True)
def combinatorial_number_system_forward(
    oriented_simplex: np.ndarray, lookup_table: np.ndarray
):
    dimension = len(oriented_simplex) - 1
    res = 0
    for l in range(dimension + 1):
        res += lookup_table[oriented_simplex[l], l + 1]
    return res


@jit(nopython=True)
def combinatorial_number_system_d1_forward(v0: int, v1: int, lookup_table: np.ndarray):
    return lookup_table[v0, 1] + lookup_table[v1, 2]


@jit(nopython=True)
def combinatorial_number_system_d2_forward(
    v0: int, v1: int, v2: int, lookup_table: np.ndarray
):
    return lookup_table[v0, 1] + lookup_table[v1, 2] + lookup_table[v2, 3]



"""
Programmer: Chris Tralie (ctralie@alumni.princeton.edu) and Luis Scoccola
Purpose: To provide a number of utility functions, including
- Quickly computing all pairs self-similarity and cross-similarity matrices
- Doing "greedy permutations" 
- Some relevant geometric examples for tests
"""
import time
import numpy as np
import scipy.sparse as sparse
from scipy.spatial import KDTree
from scipy.sparse import csr_matrix
from scipy.sparse.csgraph import shortest_path
from numba import jit
from .combinatorial import (
    combinatorial_number_system_forward,
    combinatorial_number_system_d1_forward,
    combinatorial_number_system_d2_forward,
    number_of_simplices_of_dimension,
)


class GeometryUtils:
    """
    Utilities for subsampling from point clouds and distance matrices.

    """

    @staticmethod
    def get_csm(X, Y):
        """
        Return the Euclidean cross-similarity matrix between the M points
        in the Mxd matrix X and the N points in the Nxd matrix Y.

        Parameters
        ----------
        X : ndarray (M, d)
            A matrix holding the coordinates of M points
        Y : ndarray (N, d)
            A matrix holding the coordinates of N points

        Returns
        -------
        D : ndarray (M, N)
            An MxN Euclidean cross-similarity matrix

        """
        C = np.sum(X**2, 1)[:, None] + np.sum(Y**2, 1)[None, :] - 2 * X.dot(Y.T)
        C[C < 0] = 0
        return np.sqrt(C)

    @staticmethod
    def get_csm_projarc(X, Y):
        """
        Return the projective arc length cross-similarity between two point
        clouds specified as points on the sphere

        Parameters
        ----------
        X : ndarray (M, d)
            A matrix holding the coordinates of M points on RP^{d-1}
        Y : ndarray (N, d)
            A matrix holding the coordinates of N points on RP^{d-1}

        Returns
        -------
        D : ndarray (M, N)
            An MxN  cross-similarity matrix

        """
        D = np.abs(X.dot(Y.T))
        D[D < -1] = -1
        D[D > 1] = 1
        D = np.arccos(np.abs(D))
        return D

    @staticmethod
    def get_ssm(X):
        return GeometryUtils.get_csm(X, X)

    # Greedy Permutations
    @staticmethod
    def get_greedy_perm_pc(X, M, csm_fn=get_csm.__func__):
        """
        A Naive O(NM) algorithm to do furthest points sampling, assuming
        the input is a point cloud specified in Euclidean space.  This saves
        computation over having compute the full distance matrix if the number
        of landmarks M << N

        Parameters
        ----------
        X : ndarray (N, d)
            An Nxd Euclidean point cloud
        M : integer
            Number of landmarks to compute
        csm_fn: function X, Y -> D
            Cross-similarity function (Euclidean by default)

        Return
        ------
        result: Dictionary
            {'Y': An Mxd array of landmarks,
            'perm': An array of indices into X of the greedy permutation
            'lambdas': Insertion radii of the landmarks
            'D': An MxN array of distances from landmarks to points in X}

        """
        # By default, takes the first point in the permutation to be the
        # first point in the point cloud, but could be random
        N = X.shape[0]
        perm = np.zeros(M, dtype=np.int64)
        lambdas = np.zeros(M)
        ds = csm_fn(X[0, :][None, :], X).flatten()
        D = np.zeros((M, N))
        D[0, :] = ds
        for i in range(1, M):
            idx = np.argmax(ds)
            perm[i] = idx
            lambdas[i] = ds[idx]
            thisds = csm_fn(X[idx, :][None, :], X).flatten()
            D[i, :] = thisds
            ds = np.minimum(ds, thisds)
        Y = X[perm, :]
        return {"Y": Y, "perm": perm, "lambdas": lambdas, "D": D}

    @staticmethod
    def get_greedy_perm_dm(D, M):
        """
        A Naive O(NM) algorithm to do furthest points sampling, assuming
        the input is a N x N distance matrix

        Parameters
        ----------
        D : ndarray (N, N)
            An N x N distance matrix
        M : integer
            Number of landmarks to compute

        Return
        ------
        result: Dictionary
            {'perm': An array of indices into X of the greedy permutation
            'lambdas': Insertion radii of the landmarks
            'DLandmarks': An MxN array of distances from landmarks to points in the point cloud}

        """
        # By default, takes the first point in the permutation to be the
        # first point in the point cloud, but could be random
        perm = np.zeros(M, dtype=np.int64)
        lambdas = np.zeros(M)
        ds = D[0, :]
        for i in range(1, M):
            idx = np.argmax(ds)
            perm[i] = idx
            lambdas[i] = ds[idx]
            ds = np.minimum(ds, D[idx, :])
        DLandmarks = D[perm, :]
        return {"perm": perm, "lambdas": lambdas, "DLandmarks": DLandmarks}

    @staticmethod
    def landmark_geodesic_distance(X, n_landmarks, n_neighbors):
        """
        Get a minmax sample of Euclidean point cloud using an approximation of the geodesic distance.
        Return the geodesic distance from the sample (landmarks) to the rest of the points.

        Parameters
        ----------
        X : ndarray(n,d)
            Point cloud of n points in d-dimensional Euclidean space.
        n_landmarks : int
            How many landmarks to sample.
        n_neighbors : int
            How many neighbors to use to compute the nearest neighbors graph used to approximate
            the geodesic distance.

        Return
        ------
        dist_landmarks_points : ndarray(n_landmarks,n)
            Geodesic distance from landmarks to the rest of the points.
            The first n_landmarks columns represent the landmarks, and are in the same order as landmarks.
            The rest of the columns represent the rest of the data points.
            The permutation that reorders the rows of X to correspond to the columns of dist_landmarks_points is
            given by perm_all_points.

        perm_all_points : ndarray(n)
            Permutation of [0, ..., n-1] which makes the rows of X correspond to the columns of dist_landmarks_points.

        """
        assert n_landmarks <= X.shape[0]
        assert n_neighbors <= X.shape[0]

        spatial_tree = KDTree(X)
        distances_nn, indices_nn = spatial_tree.query(X, k=n_neighbors)
        # https://github.com/scikit-learn/scikit-learn/blob/364c77e047ca08a95862becf40a04fe9d4cd2c98/sklearn/neighbors/_base.py#L997
        n_queries = X.shape[0]
        n_nonzero = n_queries * n_neighbors
        indptr = np.arange(0, n_nonzero + 1, n_neighbors)
        kneighbors_graph = csr_matrix(
            (distances_nn.ravel(), indices_nn.ravel(), indptr),
            shape=(n_queries, n_queries),
        )

        # furthest point sampling
        n_points = X.shape[0]
        perm = np.zeros(n_landmarks, dtype=np.int64)
        lambdas = np.zeros(n_landmarks)
        ds = shortest_path(kneighbors_graph, indices=0, directed=False)
        D = np.zeros((n_landmarks, n_points))
        D[0, :] = ds
        for i in range(1, n_landmarks):
            idx = np.argmax(ds)
            perm[i] = idx
            lambdas[i] = ds[idx]
            thisds = shortest_path(kneighbors_graph, indices=idx, directed=False)
            D[i, :] = thisds
            ds = np.minimum(ds, thisds)

        perm_rest_points = np.setdiff1d(
            np.arange(0, n_points, dtype=int), perm, assume_unique=True
        )
        perm_all_points = np.concatenate((perm, perm_rest_points))

        dist_landmarks_points = D[:, perm_all_points]

        return dist_landmarks_points, perm_all_points


class CohomologyUtils:
    @staticmethod
    def lift_to_integer_cocycle(cocycle, prime):
        """
        Lift the given cocycle with values in a prime field to a cocycle with integer coefficients.

        Parameters
        ----------
        cocycle : ndarray(K, n, dtype=int)
            Cocycle to be lifted to integer coefficients.

        Note
        ----
        This routine modifies the input cocycle.

        Returns
        -------
        cocycle : ndarray(K, n, dtype=int)
            Cocycle with same support as input cocycle and integer coefficients.

        """
        cocycle[cocycle[:, -1] > (prime - 1) / 2, -1] -= prime
        return cocycle

    @staticmethod
    def reindex_cocycles(cocycles, idx_land, N):
        """
        Convert the indices of a set of cocycles to be relative
        to a list of indices in a greedy permutation

        Parameters
        ----------
        cocycles: list of list of ndarray
            The cocycles
        idx_land: ndarray(M, dtype=int)
            Indices of the landmarks in the greedy permutation, with
            respect to all points
        N: int
            Number of total points

        """
        idx_map = -1 * np.ones(N, dtype=int)
        idx_map[idx_land] = np.arange(idx_land.size)
        for ck in cocycles:
            for c in ck:
                c[:, 0:-1] = idx_map[c[:, 0:-1]]

    @staticmethod
    def parity(permutation):
        """Compute the parity of a permutation"""
        permutation_length = len(permutation)
        elements_seen = [False for _ in range(permutation_length)]
        cycles = 0
        for index, already_seen in enumerate(elements_seen):
            if already_seen:
                continue
            cycles += 1
            current = index
            while not elements_seen[current]:
                elements_seen[current] = True
                current = permutation[current]
        return 1 if (permutation_length - cycles) % 2 == 0 else -1

    @staticmethod
    def order_simplex(unordered_simplex):
        ordering_permutation = np.argsort(unordered_simplex)
        sign = CohomologyUtils.parity(ordering_permutation)
        simplex = tuple(unordered_simplex[ordering_permutation])
        return simplex, sign

    @staticmethod
    def sparse_cocycle_to_vector(sparse_cocycle, lookup_table, n_vertices, dtype):
        dimension = sparse_cocycle.shape[1] - 2
        n_simplices = number_of_simplices_of_dimension(
            dimension, n_vertices, lookup_table
        )
        cocycle_as_vector = np.zeros((n_simplices,), dtype=dtype)
        for entry in sparse_cocycle:
            value = entry[-1]
            unordered_simplex = np.array(entry[:-1], dtype=int)
            ordered_simplex, sign = CohomologyUtils.order_simplex(unordered_simplex)
            simplex_index = combinatorial_number_system_forward(
                ordered_simplex, lookup_table
            )
            cocycle_as_vector[simplex_index] = value * sign
        return cocycle_as_vector

    @staticmethod
    def make_delta0(dist_mat: np.ndarray, threshold: float, lookup_table: np.ndarray):
        n_points = dist_mat.shape[0]
        n_edges = (n_points * (n_points - 1)) // 2

        max_n_entries = 2 * n_edges
        rows = np.empty((max_n_entries,), dtype=np.int32)
        columns = np.empty((max_n_entries,), dtype=np.int32)
        values = np.empty((max_n_entries,), dtype=float)

        @jit(fastmath=True, nopython=True)
        def _delta0_get_row_columns_values(
            dist_mat: np.ndarray,
            threshold: float,
            lookup_table: np.ndarray,
            n_points: int,
            rows: np.ndarray,
            columns: np.ndarray,
            values: np.ndarray,
        ):
            n_entries = 0
            for i in range(n_points):
                for j in range(i + 1, n_points):
                    if dist_mat[i, j] < threshold:
                        row_index = combinatorial_number_system_d1_forward(
                            i, j, lookup_table
                        )
                        rows[n_entries] = row_index
                        columns[n_entries] = i
                        values[n_entries] = 1
                        n_entries += 1

                        rows[n_entries] = row_index
                        columns[n_entries] = j
                        values[n_entries] = -1
                        n_entries += 1
            return n_entries

        n_entries = _delta0_get_row_columns_values(
            dist_mat, threshold, lookup_table, n_points, rows, columns, values
        )

        return sparse.csr_array(
            (values[:n_entries], (rows[:n_entries], columns[:n_entries])),
            shape=(n_edges, n_points),
        )

    @staticmethod
    def make_delta1(dist_mat: np.ndarray, threshold: float, lookup_table: np.ndarray):
        n_points = dist_mat.shape[0]
        n_edges = (n_points * (n_points - 1)) // 2
        n_faces = number_of_simplices_of_dimension(2, n_points, lookup_table)

        max_n_entries = 3 * n_faces
        rows = np.empty((max_n_entries,), dtype=np.int32)
        columns = np.empty((max_n_entries,), dtype=np.int32)
        values = np.empty((max_n_entries,), dtype=float)

        @jit(fastmath=True, nopython=True)
        def _delta1_get_row_columns_values(
            dist_mat: np.ndarray,
            threshold: float,
            lookup_table: np.ndarray,
            n_points: int,
            rows: np.ndarray,
            columns: np.ndarray,
            values: np.ndarray,
        ):
            n_entries = 0
            for i in range(n_points):
                for j in range(i + 1, n_points):
                    if dist_mat[i, j] < threshold:
                        for k in range(j + 1, n_points):
                            if (
                                dist_mat[i, k] < threshold
                                and dist_mat[j, k] < threshold
                            ):
                                row_index = combinatorial_number_system_d2_forward(
                                    i, j, k, lookup_table
                                )
                                column_index_ij = (
                                    combinatorial_number_system_d1_forward(
                                        i, j, lookup_table
                                    )
                                )
                                column_index_jk = (
                                    combinatorial_number_system_d1_forward(
                                        j, k, lookup_table
                                    )
                                )
                                column_index_ik = (
                                    combinatorial_number_system_d1_forward(
                                        i, k, lookup_table
                                    )
                                )
                                rows[n_entries] = row_index
                                columns[n_entries] = column_index_ij
                                values[n_entries] = 1
                                n_entries += 1

                                rows[n_entries] = row_index
                                columns[n_entries] = column_index_jk
                                values[n_entries] = 1
                                n_entries += 1

                                rows[n_entries] = row_index
                                columns[n_entries] = column_index_ik
                                values[n_entries] = -1
                                n_entries += 1

            return n_entries

        n_entries = _delta1_get_row_columns_values(
            dist_mat, threshold, lookup_table, n_points, rows, columns, values
        )

        # print("delta 1")
        # print(n_entries)
        # print(max_n_entries)

        return sparse.csr_array(
            (values[:n_entries], (rows[:n_entries], columns[:n_entries])),
            shape=(n_faces, n_edges),
        )

    @staticmethod
    def make_delta1_compact(
        dist_mat: np.ndarray, threshold: float, lookup_table: np.ndarray
    ):
        """
        Like [make_delta1] but it does not use the combinatorial number system for the rows
        (i.e., for the 2-simplices, and only has a row per 2-simplex *that exists in the filtration at that point*.
        This is used to solve the integer linear problem since, in that case, we do not need to index the 2-simplices
        so only having the ones that exist in the filtration at that point results in a matrix with potentially much
        fewer rows.
        """
        n_points = dist_mat.shape[0]
        n_edges = (n_points * (n_points - 1)) // 2
        n_faces = number_of_simplices_of_dimension(2, n_points, lookup_table)

        max_n_entries = 3 * n_faces
        rows = np.empty((max_n_entries,), dtype=np.int32)
        columns = np.empty((max_n_entries,), dtype=np.int32)
        values = np.empty((max_n_entries,), dtype=float)

        @jit(fastmath=True, nopython=True)
        def _delta1_get_row_columns_values(
            dist_mat: np.ndarray,
            threshold: float,
            lookup_table: np.ndarray,
            n_points: int,
            rows: np.ndarray,
            columns: np.ndarray,
            values: np.ndarray,
        ):
            n_entries = 0
            n_actual_faces = 0
            for i in range(n_points):
                for j in range(i + 1, n_points):
                    if dist_mat[i, j] < threshold:
                        for k in range(j + 1, n_points):
                            if (
                                dist_mat[i, k] < threshold
                                and dist_mat[j, k] < threshold
                            ):
                                # row_index = combinatorial_number_system_d2_forward(
                                #    i, j, k, lookup_table
                                # )
                                row_index = n_actual_faces
                                column_index_ij = (
                                    combinatorial_number_system_d1_forward(
                                        i, j, lookup_table
                                    )
                                )
                                column_index_jk = (
                                    combinatorial_number_system_d1_forward(
                                        j, k, lookup_table
                                    )
                                )
                                column_index_ik = (
                                    combinatorial_number_system_d1_forward(
                                        i, k, lookup_table
                                    )
                                )
                                rows[n_entries] = row_index
                                columns[n_entries] = column_index_ij
                                values[n_entries] = 1
                                n_entries += 1

                                rows[n_entries] = row_index
                                columns[n_entries] = column_index_jk
                                values[n_entries] = 1
                                n_entries += 1

                                rows[n_entries] = row_index
                                columns[n_entries] = column_index_ik
                                values[n_entries] = -1
                                n_entries += 1

                                n_actual_faces += 1
            return n_entries, n_actual_faces

        n_entries, n_actual_faces = _delta1_get_row_columns_values(
            dist_mat, threshold, lookup_table, n_points, rows, columns, values
        )

        # print("delta 1")
        # print(n_entries)
        # print(max_n_entries)

        return sparse.csr_array(
            (values[:n_entries], (rows[:n_entries], columns[:n_entries])),
            shape=(n_actual_faces, n_edges),
        )

    @staticmethod
    def make_delta2_compact(
        dist_mat: np.ndarray, threshold: float, lookup_table: np.ndarray
    ):
        """
        Like [make_delta1_compact] but for delta2.
        """
        n_points = dist_mat.shape[0]
        n_faces = number_of_simplices_of_dimension(2, n_points, lookup_table)
        n_three_simplices = number_of_simplices_of_dimension(3, n_points, lookup_table)

        max_n_entries = 4 * n_three_simplices
        rows = np.empty((max_n_entries,), dtype=np.int32)
        columns = np.empty((max_n_entries,), dtype=np.int32)
        values = np.empty((max_n_entries,), dtype=float)

        @jit(fastmath=True, nopython=True)
        def _delta2_get_row_columns_values(
            dist_mat: np.ndarray,
            threshold: float,
            lookup_table: np.ndarray,
            n_points: int,
            rows: np.ndarray,
            columns: np.ndarray,
            values: np.ndarray,
        ):
            n_entries = 0
            n_actual_three_simplices = 0
            for i in range(n_points):
                for j in range(i + 1, n_points):
                    if dist_mat[i, j] < threshold:
                        for k in range(j + 1, n_points):
                            if (
                                dist_mat[i, k] < threshold
                                and dist_mat[j, k] < threshold
                            ):
                                for l in range(k + 1, n_points):
                                    if (
                                        dist_mat[i, l] < threshold
                                        and dist_mat[j, l] < threshold
                                        and dist_mat[k, l] < threshold
                                    ):
                                        row_index = n_actual_three_simplices

                                        column_index_ijk = (
                                            combinatorial_number_system_d2_forward(
                                                i, j, k, lookup_table
                                            )
                                        )
                                        column_index_ijl = (
                                            combinatorial_number_system_d2_forward(
                                                i, j, l, lookup_table
                                            )
                                        )
                                        column_index_ikl = (
                                            combinatorial_number_system_d2_forward(
                                                i, k, l, lookup_table
                                            )
                                        )
                                        column_index_jkl = (
                                            combinatorial_number_system_d2_forward(
                                                j, k, l, lookup_table
                                            )
                                        )

                                        rows[n_entries] = row_index
                                        columns[n_entries] = column_index_ijk
                                        values[n_entries] = 1
                                        n_entries += 1

                                        rows[n_entries] = row_index
                                        columns[n_entries] = column_index_ijl
                                        values[n_entries] = -1
                                        n_entries += 1

                                        rows[n_entries] = row_index
                                        columns[n_entries] = column_index_ikl
                                        values[n_entries] = 1
                                        n_entries += 1

                                        rows[n_entries] = row_index
                                        columns[n_entries] = column_index_jkl
                                        values[n_entries] = -1
                                        n_entries += 1

                                        n_actual_three_simplices += 1
            return n_entries, n_actual_three_simplices

        n_entries, n_actual_three_simplices = _delta2_get_row_columns_values(
            dist_mat, threshold, lookup_table, n_points, rows, columns, values
        )

        # print("delta 1")
        # print(n_entries)
        # print(max_n_entries)

        return sparse.csr_array(
            (values[:n_entries], (rows[:n_entries], columns[:n_entries])),
            shape=(n_actual_three_simplices, n_faces),
        )


class PartUnity:
    """
    Partitions of unity subordinate to open ball covers using
    standard bump functions.

    """

    @staticmethod
    def linear(ds, r_cover):
        """
        Linear partition of unity.

        Parameters
        ----------
        ds: ndarray(n)
            Some subset of distances between landmarks and
            data points
        r_cover: float
            Covering radius

        Returns
        -------
        varphi: ndarray(n)
            The bump function

        """
        return r_cover - ds

    @staticmethod
    def quadratic(ds, r_cover):
        """
        Quadratic partition of unity.

        Parameters
        ----------
        ds: ndarray(n)
            Some subset of distances between landmarks and
            data points
        r_cover: float
            Covering radius

        Returns
        -------
        varphi: ndarray(n)
            The bump function

        """
        return (r_cover - ds) ** 2

    @staticmethod
    def exp(ds, r_cover):
        """
        Exponential partition of unity.

        Parameters
        ----------
        ds: ndarray(n)
            Some subset of distances between landmarks and
            data points
        r_cover: float
            Covering radius

        Returns
        -------
        varphi: ndarray(n)
            The bump function

        """
        return np.exp(r_cover**2 / (ds**2 - r_cover**2))


class EquivariantPCA:
    @staticmethod
    def ppca(class_map, proj_dim, projective_dim_red_mode="one-by-one", verbose=False):
        """
        Principal Projective Component Analysis (Jose Perea 2017)

        Parameters
        ----------
        class_map : ndarray (N, d)
            For all N points of the dataset, membership weights to
            d different classes are the coordinates
        proj_dim : integer
            The dimension of the projective space onto which to project
        projective_dim_red_mode : string
            Either "one-by-one", "exponential", or "direct". How to perform equivariant
            dimensionality reduction. "exponential" usually works best, being fast
            without compromising quality.
        verbose : boolean
            Whether to print information during iterations

        Returns
        -------
        {'variance': ndarray(N-1)
            The variance captured by each dimension
        'X': ndarray(N, proj_dim+1)
            The projective coordinates
        }

        """
        if verbose:
            print(
                "Doing ppca on %i points in %i dimensions down to %i dimensions"
                % (class_map.shape[0], class_map.shape[1], proj_dim)
            )

        assert projective_dim_red_mode in ["direct", "exponential", "one-by-one"]

        X = class_map.T
        variance = np.zeros(X.shape[0] - 1)
        n_dim = class_map.shape[1]

        def _one_step_linear_reduction(X, dims_to_keep):
            try:
                _, U = np.linalg.eigh(X.dot(np.conjugate(X).T))
                U = np.fliplr(U)
            except:
                U = np.eye(X.shape[0])
            Y = (np.conjugate(U).T).dot(X)
            Y = Y[:dims_to_keep, :]
            X = Y / np.linalg.norm(Y, axis=0)[None, :]
            return X

        total_dims_to_keep = proj_dim + 1

        mode = projective_dim_red_mode
        if mode == "direct":
            XRet = _one_step_linear_reduction(X, total_dims_to_keep)
        elif mode == "exponential":
            to_keep_this_iter = (n_dim - total_dims_to_keep) // 2
            while to_keep_this_iter > 0:
                X = _one_step_linear_reduction(
                    X, total_dims_to_keep + to_keep_this_iter
                )
                to_keep_this_iter = to_keep_this_iter // 2
            if X.shape[0] > total_dims_to_keep:
                X = _one_step_linear_reduction(X, total_dims_to_keep)
            XRet = X

        elif mode == "one-by-one":
            tic = time.time()
            # Projective dimensionality reduction : Main Loop
            XRet = None
            for i in range(n_dim - 1):
                if i == n_dim - proj_dim - 1:
                    XRet = X
                try:
                    _, U = np.linalg.eigh(X.dot(np.conjugate(X).T))
                    U = np.fliplr(U)
                    # U, _, _ = np.linalg.svd(X)
                except:
                    U = np.eye(X.shape[0])
                variance[-i - 1] = np.mean(
                    (np.pi / 2 - np.real(np.arccos(np.abs(U[:, -1][None, :].dot(X)))))
                    ** 2
                )
                Y = (np.conjugate(U).T).dot(X)
                # y = np.array(Y[-1, :])
                Y = Y[0:-1, :]
                # X = Y / np.sqrt(1 - np.abs(y) ** 2)[None, :]
                X = Y / np.linalg.norm(Y, axis=0)[None, :]
            if verbose:
                print("Elapsed time ppca: %.3g" % (time.time() - tic))

        # Return the variance and the projective coordinates
        return {"variance": variance, "X": XRet.T}


class GeometryExamples:
    """
    Finite samples from topologically nontrivial spaces.

    """

    # TODO: These could belong to tdasets, but we'll keep them here for now

    @staticmethod
    def line_patches(dim, n_angles, n_offsets, sigma):
        """
        Sample a set of line segments, as witnessed by square patches

        Parameters
        ----------
        dim: int
            Patches will be dim x dim
        n_angles: int
            Number of angles to sweep between 0 and pi
        n_offsets: int
            Number of offsets to sweep from the origin to the edge of the patch
        sigma: float
            The blur parameter.  Higher sigma is more blur

        Returns
        -------
        ndarray(n_angles*n_offsets, dim*dim)
            An array of all of the patches raveled into dim*dim dimensional Euclidean space

        """
        N = n_angles * n_offsets
        P = np.zeros((N, dim * dim))
        thetas = np.linspace(0, np.pi, n_angles + 1)[0:n_angles]
        # ps = np.linspace(-0.5*np.sqrt(2), 0.5*np.sqrt(2), n_offsets)
        ps = np.linspace(-1, 1, n_offsets)
        idx = 0
        [Y, X] = np.meshgrid(np.linspace(-0.5, 0.5, dim), np.linspace(-0.5, 0.5, dim))
        for i in range(n_angles):
            c = np.cos(thetas[i])
            s = np.sin(thetas[i])
            for j in range(n_offsets):
                patch = X * c + Y * s + ps[j]
                patch = np.exp(-(patch**2) / sigma**2)
                P[idx, :] = patch.flatten()
                idx += 1
        return P

    @staticmethod
    def moving_dot(sqrt_num_images, sigma=3, dim=10):
        """
        Sample a set of (smoothened) dots on the plane, as witnessed by square patches

        Parameters
        ----------
        sqrt_num_images : int
            The output will consist of sqrt_num_images**2 patches.

        sigma : float
            The blur parameter. Higher sigma is more blur. Default is 3.

        dim : int
            Patches will be dim x dim. Default is 10.

        Returns
        -------
        ndarray(sqrt_num_images*sqrt_num_images, dim*dim)
            An array of all of the patches raveled into dim*dim dimensional Euclidean space

        """

        def _gkern(l=5, mu=0, sig=1.0):
            ax = np.linspace(-(l - 1) / 2.0, (l - 1) / 2.0, l)
            gauss_x = np.exp(-0.5 * np.square(ax - mu[0]) / np.square(sig))
            gauss_y = np.exp(-0.5 * np.square(ax - mu[1]) / np.square(sig))
            kernel = np.outer(gauss_x, gauss_y)
            return kernel

        dim = 10
        P = np.zeros((sqrt_num_images**2, dim * dim))
        bound = 15
        xs = bound * np.power(np.linspace(-1, 1, sqrt_num_images), 3)
        # xs = bound * np.linspace(-1,1,sqrt_num_images)
        ys = -xs
        i = 0
        for x in xs:
            for y in ys:
                P[i] = _gkern(l=dim, mu=np.array([x, y]), sig=sigma).flatten()
                i += 1
        return P

    @staticmethod
    def rp2_metric(n_samples, seed=None):
        """
        Return a distance matrix of points on the projective plane
        obtained by identifying antipodal Gaussian random samples
        of a sphere

        Parameters
        ----------
        n_samples : int
            Number of random samples on the projective plane
        seed: int
            Seed to use.  If omitted, use the number of samples as a seed

        Returns
        -------
        ndarray(n_samples, 3)
            Original points on the sphere
        ndarray(n_samples, n_samples)
            Distance matrix of rp2

        """
        if seed is None:
            seed = n_samples
        np.random.seed(seed)
        X = np.random.randn(n_samples, 3)
        X = X / np.sqrt(np.sum(X**2, 1))[:, None]
        return X, GeometryUtils.get_csm_projarc(X, X)

    @staticmethod
    def torus_3d(n_samples, R, r, seed=None):
        """
        Return points sampled on a 3D torus

        Parameters
        ----------
        n_samples : int
            Number of random samples on the torus
        R: float
            Outer radius
        r: float
            Inner radius
        seed: int
            Seed to use.  If omitted, use the number of samples as a seed

        Returns
        -------
        X: ndarray(n_samples, 4)
            3D torus samples

        """
        if seed is None:
            seed = n_samples
        np.random.seed(seed)
        X = np.zeros((n_samples, 3))
        s = np.random.rand(n_samples) * 2 * np.pi
        t = np.random.rand(n_samples) * 2 * np.pi
        X[:, 0] = (R + r * np.cos(s)) * np.cos(t)
        X[:, 1] = (R + r * np.cos(s)) * np.sin(t)
        X[:, 2] = r * np.sin(s)
        return X

    @staticmethod
    def klein_bottle_4d(n_samples, R, r, seed=None):
        """
        Return samples on a klein bottle in 4D

        Parameters
        ----------
        n_samples : int
            Number of random samples on the projective plane
        R: float
            Outer radius
        r: float
            Inner radius
        seed: int
            Seed to use.  If omitted, use the number of samples as a seed

        Returns
        -------
        X: ndarray(n_samples, 4)
            4D klein bottle samples

        """
        if seed is None:
            seed = n_samples
        np.random.seed(seed)
        theta = np.random.rand(n_samples) * 2 * np.pi
        phi = np.random.rand(n_samples) * 2 * np.pi
        X = np.zeros((n_samples, 4))
        X[:, 0] = (R + r * np.cos(theta)) * np.cos(phi)
        X[:, 1] = (R + r * np.cos(theta)) * np.sin(phi)
        X[:, 2] = r * np.sin(theta) * np.cos(phi / 2)
        X[:, 3] = r * np.sin(theta) * np.sin(phi / 2)
        return X

    @staticmethod
    def genus_two_surface():
        """
        Return samples on a genus two surface in 3D

        Returns
        -------
        X: ndarray(n_samples, 3)
            3D surface samples

        """

        R2 = 9
        R = 5
        r = 2
        Ns = 80
        Nt = 120
        N = Ns * Nt
        Y = np.zeros((N, 3))
        s = np.linspace(0, 2 * np.pi, Ns, endpoint=False)
        t = np.linspace(0, 2 * np.pi, Nt, endpoint=False)
        st = np.array([[x, y] for x in s for y in t])
        s = st[:, 0]
        t = st[:, 1]
        Y[:, 0] = (R + r * np.cos(s)) * np.cos(t)
        Y[:, 1] = (R + r * np.cos(s)) * np.sin(t)
        Y[:, 2] = r * np.sin(s)

        Z = np.zeros((N, 3))
        Z[:, 0] = R2 + (R + r * np.cos(s)) * np.cos(t)
        Z[:, 1] = (R + r * np.cos(s)) * np.sin(t)
        Z[:, 2] = r * np.sin(s)

        Y = Y[(Y[:, 0] <= 4.5)]
        Z = Z[(Z[:, 0] >= 4.5)]

        return np.concatenate((Y, Z), axis=0)

    @staticmethod
    def trefoil(n_samples=2500, horizontal_width=6, noisy=True):
        """
        Samples on a trefoil in 3D

        Parameters
        ----------
        n_samples : int
            Number of random samples.

        Returns
        -------
        X: ndarray(n_samples, 3)
            3D trefoil samples

        """

        if noisy:
            np.random.seed(0)
            u = 4 * np.pi * np.random.rand(n_samples)
            v = 2 * np.pi * np.random.rand(n_samples)
            X = np.zeros((n_samples, 3))
            X[:, 0] = np.cos(u) * np.cos(v) + horizontal_width * np.cos(u) * (
                1.5 + np.sin(1.5 * u) / 2
            )
            X[:, 1] = np.sin(u) * np.cos(v) + horizontal_width * np.sin(u) * (
                1.5 + np.sin(1.5 * u) / 2
            )
            X[:, 2] = np.sin(v) + 4 * np.cos(1.5 * u)
        else:
            np.random.seed(0)
            u = 4 * np.pi * np.linspace(0, 1, n_samples, endpoint=False)
            X = np.zeros((n_samples, 3))
            X[:, 0] = np.cos(u) + horizontal_width * np.cos(u) * (
                1.5 + np.sin(1.5 * u) / 2
            )
            X[:, 1] = np.sin(u) + horizontal_width * np.sin(u) * (
                1.5 + np.sin(1.5 * u) / 2
            )
            X[:, 2] = 4 * np.cos(1.5 * u)

        return X

    @staticmethod
    def bullseye():
        """
        Samples on three concentric noisy circles in 2D.

        Returns
        -------
        X: ndarray(n_samples, 2)
            2D circles samples

        """

        N = 200
        sample_interval = np.linspace(0, 2 * np.pi, N, endpoint=False)
        c1 = np.array([np.sin(sample_interval), np.cos(sample_interval)]).T
        c2 = np.array([2 * np.sin(sample_interval), 2 * np.cos(sample_interval)]).T
        c3 = np.array([3 * np.sin(sample_interval), 3 * np.cos(sample_interval)]).T
        X = np.vstack((c1, c2, c3))

        np.random.seed(0)
        eps = 0.1
        X += (np.random.random(X.shape) - 0.5) * eps

        return X

    @staticmethod
    def three_circles():
        """
        Samples on two circles of different radii in 2D.

        Returns
        -------
        X: ndarray(n_samples, 2)
            2D circles samples

        """

        sample_interval_small_circle = np.linspace(0, 2 * np.pi, 50, endpoint=False)
        small_circle = np.array(
            [np.sin(sample_interval_small_circle), np.cos(sample_interval_small_circle)]
        ).T
        sample_interval_big_circle = np.linspace(0, 2 * np.pi, 500, endpoint=False)
        big_circle1 = np.array(
            [
                2 * np.sin(sample_interval_big_circle),
                2 * np.cos(sample_interval_big_circle),
            ]
        ).T + np.array([4, 0])
        big_circle2 = np.array(
            [
                2.5 * np.sin(sample_interval_big_circle),
                2.5 * np.cos(sample_interval_big_circle),
            ]
        ).T + np.array([-4, 0])
        X = np.vstack((small_circle, big_circle1, big_circle2))

        np.random.seed(0)
        eps = 0.3
        X += (np.random.random(X.shape) - 0.5) * eps

        return X

    @staticmethod
    def sphere(n_samples):
        """
        Samples on a 2-sphere in 3D.

        Returns
        -------
        X: ndarray(n_samples, 3)
            3D sphere samples

        """
        np.random.seed(0)
        data = 2 * np.random.random_sample((n_samples, 3)) - 1
        return data / np.linalg.norm(data, axis=1)[:, np.newaxis]

    @staticmethod
    def noisy_circle(n_samples, noise_size=0.2, seed=0):
        """
        Samples on a circle in 2D.

        Parameters
        ----------
        n_samples : int
            Number of points on the circle to sample.

        noise_size : float
            Maximum perturbation per sample. Must be between 0 and 1.
        
        seed : float
            Seed for random number generator.

        Returns
        -------
        X: ndarray(n_samples, 2)
            2D circle samples

        """
        np.random.seed(seed)
        X = np.random.random((n_samples, 2)) - 0.5
        return (
            X / np.linalg.norm(X, axis=1).reshape((n_samples, 1))
            + (np.random.random((n_samples, 2)) - 0.5) * noise_size
        )

    @staticmethod
    def moore_space_distance_matrix(rough_n_points=2000, prime=3):
        """
        Distance matrix of a sample of the Moore space M(1,Z/prime).

        Parameters
        ----------
        rough_n_samples : int
            The function returns a distance matrix with approximately n = rough_n_points * (pi/4) points.

        prime : int
            The prime associated with the Moore space.

        Returns
        -------
        ndarray(n, n)
            Distance matrix of points on the Moore space M(1,Z/prime).

        """
        np.random.seed(0)
        X = (np.random.random((rough_n_points, 2)) - 0.5) * 2
        X = X[np.linalg.norm(X, axis=1) <= 1]
        q = prime

        def _rot_mat(theta):
            c, s = np.cos(theta), np.sin(theta)
            return np.array(((c, -s), (s, c)))

        R = _rot_mat((2 * np.pi) / q)

        n_points = X.shape[0]
        dist_mat = np.zeros((n_points, n_points))

        @jit(fastmath=True, nopython=True)
        def _fill_dist_mat(X, dist_mat, rot_mat, prime):
            for i, x in enumerate(X):
                for j, y in enumerate(X):
                    proj_x_to_boundary = x / np.linalg.norm(x)

                    dist_mat[i, j] = min(
                        # stay inside disk
                        np.linalg.norm(x - y),
                        # go to boundary and then to y
                        np.linalg.norm(x - proj_x_to_boundary)
                        + min(
                            [
                                np.linalg.norm(
                                    np.linalg.matrix_power(rot_mat, i)
                                    @ proj_x_to_boundary
                                    - y
                                )
                                for i in range(prime)
                            ]
                        ),
                    )

        _fill_dist_mat(X, dist_mat, R, prime)
        return dist_mat, X


class CircleMapUtils:
    """

    Utilities for adding, rotating, and plotting circle-valued maps.

    """

    @staticmethod
    def to_sinebow(circle_map):
        """
        Given a circle map construct an array of the same shape that can be fed as color in a matplotlib
        scatterplot to simulate the sinebow colormap.

        Parameters
        ----------
        circle_map: ndarray
            A numpy array of numbers between 0 and 2pi representing
            points on the circle.

        Returns
        -------
        ndarray
            A numpy array of floats to be used as color in a matplotlib scatterplot.

        """
        h = np.mod(circle_map / (2 * np.pi) + 0.5, 1)
        f = lambda x: np.sin(np.pi * x) ** 2
        return np.stack([f(3 / 6 - h), f(5 / 6 - h), f(7 / 6 - h)], -1)

    @staticmethod
    def center(circle_map):
        """
        Rotationally offset a circle-valued map so that most
        of the points map to the center of the circle (i.e., pi).

        Parameters
        ----------
        circle_map: ndarray
            A numpy array of numbers between 0 and 2pi representing
            points on the circle.

        Returns
        -------
        ndarray
            A numpy array of numbers between 0 and 2pi representing
            the rotation of the given points by the given offset.
        """
        bins = 50
        vals, ticks = np.histogram(circle_map, bins=bins)
        centered = ((circle_map - ticks[np.argmax(vals)]) + np.pi) % (2 * np.pi)
        return centered

    @staticmethod
    def offset(circle_map, offset):
        """
        Rotationally offset a circle-valued map.

        ----------
        circle_map: ndarray
            A numpy array of numbers between 0 and 2pi representing
            points on the circle.

        offset: float
            A number between 0 and 1 representing a rotational offset.

        Returns
        -------
        ndarray
            A numpy array of numbers between 0 and 2pi representing
            the rotation of the given points by the given offset.

        """

        return (circle_map + offset * (2 * np.pi)) % (2 * np.pi)

    @staticmethod
    def linear_combination(circle_maps, linear_combination_matrix):
        """
        Given k circle-valued maps on a dataset with n points and an l x k
        matrix with integer coefficients, return the l linear combinations of the
        given circle-valued maps induced by the given matrix.

        Parameters
        ----------
        circle_maps: list or ndarray(k, n, dtype=float)
            A numpy array with rows containing n points in the circle represented as
            floats between 0 and 2pi.

        linear_combination_matrix: list or ndarray(l, k, dtype=int)
            A numpy array encoding l integer linear combinations of the given k
            circle-valued maps.

        Returns
        -------
        ndarray(l, n, dtype=float)
            A numpy array with rows containing n points in the circle representing
            the l linear combinations of the given k circle-valued maps.
        """

        if not isinstance(circle_maps, np.ndarray):
            circle_maps = np.array(circle_maps)
        assert len(circle_maps.shape) == 2
        if not isinstance(linear_combination_matrix, np.ndarray):
            linear_combination_matrix = np.array(linear_combination_matrix)
        assert (
            len(linear_combination_matrix.shape) == 2
            or len(linear_combination_matrix.shape) == 1
        )

        if len(linear_combination_matrix.shape) == 2:
            assert circle_maps.shape[0] == linear_combination_matrix.shape[1]
            return (linear_combination_matrix @ circle_maps) % (2 * np.pi)
        else:
            assert circle_maps.shape[0] == linear_combination_matrix.shape[0]
            return (
                (np.array([linear_combination_matrix]) @ circle_maps) % (2 * np.pi)
            )[0]

    @staticmethod
    def levelset_coloring(circle_map, n_levelsets=4, smoothing=0.25):
        """
        Given points on the circle and a number of levelsets subdivide the
        circle into the given number of levelsets and return a smoothened
        membership function to the levelsets. This is useful for coloring a
        dataset X according to a circle-valued map X -> S^1.

        Parameters
        ----------
        circle_map: ndarray
            A numpy array of numbers between 0 and 2pi representing
            points on the circle.

        n_levelset: int, optional, default is 4
            Number of levelsets to evenly cover the circle.

        smoothing: float, optional, default is 0.25
            How much to smoothen the membership function

        Returns
        -------
        ndarray
            The smoothened membership function of each of the given
            points in the circle.
        """
        assert isinstance(n_levelsets, int)
        assert n_levelsets > 0
        n_levelsets *= 2
        colors = circle_map / (2 * np.pi)
        # transition should be between 0 (very fast) and 1 (slow)
        if smoothing == 0:
            return np.array([np.floor(c * n_levelsets) % 2 for c in colors])
        k = smoothing

        def sigmoid(x):
            x = (x - 0.5) * 2
            s = 1 / (1 + np.exp(-x / k))
            return s

        def triangle(y):
            return y if y < 1 else 2 - y

        return np.array([sigmoid(triangle((c * n_levelsets) % 2)) for c in colors])


class ProjectiveMapUtils:
    """
    Utilities for manipulating projective space-valued maps.

    """

    @staticmethod
    def get_stereo_proj_codim1(pX, u=np.array([])):
        """
        Do a projective stereographic projection

        Parameters
        ----------
        pX: ndarray(N, d)
            A collection of N points in d dimensions
        u: ndarray(d)
            A unit vector representing the north pole
            for the stereographic projection

        Returns
        -------
        S: ndarray(N, d-1)
            The stereographically projected coordinates

        """

        X = pX.T
        # Put points all on the same hemisphere
        if u.size == 0:
            _, U = np.linalg.eigh(X.dot(X.T))
            u = U[:, 0]
        XX = ProjectiveMapUtils.rotmat(u).dot(X)
        ind = XX[-1, :] < 0
        XX[:, ind] *= -1
        # Do stereographic projection
        S = XX[0:-1, :] / (1 + XX[-1, :])[None, :]
        return S.T

    @staticmethod
    def circle_to_3dnorthpole(x):
        """
        Convert a point selected on the circle to a 3D
        unit vector on the upper hemisphere

        Parameters
        ----------
        x: ndarray(2)
            Selected point in the circle

        Returns
        -------
        x: ndarray(2)
            Selected point in the circle (possibly clipped to the circle)
        u: ndarray(3)
            Unit vector on the upper hemisphere

        """
        magSqr = np.sum(x**2)
        if magSqr > 1:
            x /= np.sqrt(magSqr)
            magSqr = 1
        u = np.zeros(3)
        u[0:2] = x
        u[2] = np.sqrt(1 - magSqr)
        return x, u

    @staticmethod
    def hopf_map(X):
        """
        Use the Hopf map to project points on the unit sphere in C^2 representing points in CP^1 to points on the unit sphere in R^3.

        Parameters
        ----------
        X : complex ndarray(n,2)
            Points on the unit sphere in C^2.

        Returns
        -------
        real ndarray(n,3)
            Points on the unit sphere in R^3.

        """
        Y = np.zeros((2 * X.shape[1], X.shape[0]))
        Y[::2, :] = np.real(X).T
        Y[1::2, :] = np.imag(X).T
        return np.array(
            [
                2 * (np.prod(Y[[0, 2], :], axis=0) + np.prod(Y[[1, 3], :], axis=0)),
                2 * (np.prod(Y[[1, 2], :], axis=0) - np.prod(Y[[0, 3], :], axis=0)),
                np.sum(Y[[0, 1], :] ** 2, axis=0) - np.sum(Y[[2, 3], :] ** 2, axis=0),
            ]
        ).T

    @staticmethod
    def stereographic_projection_hemispheres(X, center_vector=None):
        """
        Project points on the unit sphere in 3D to two disks in 2D corresponding to each hemisphere.

        Parameters
        ----------
        X : ndarray(n,3)
            Points on the unit sphere in 3D

        Returns
        -------
        ndarray(n,2)

        """

        def _stereo(v):
            return v[:, :-1] / (1 - v[:, -1])[:, None]

        n = X.shape[1]
        if center_vector is None:
            center_vector = np.zeros((n))
            center_vector[-1] = 1
        centering_rotation = ProjectiveMapUtils.rotmat(center_vector)
        X_ = X @ centering_rotation.T
        e1 = np.zeros((n - 1))
        e1[0] = 1
        res = np.zeros((X_.shape[0], n - 1))
        res[X_[:, -1] < 0, :] = _stereo(X_[X_[:, -1] < 0, :])
        Y = X_[X_[:, -1] >= 0, :]
        Y[:, -1] *= -1
        res[X_[:, -1] >= 0, :] = _stereo(Y) + 2.5 * e1
        return res

    @staticmethod
    def rotmat(a, b=np.array([])):
        """
        Construct a d x d rotation matrix that rotates
        a vector a so that it coincides with a vector b

        Parameters
        ----------
        a : ndarray (d)
            A d-dimensional vector that should be rotated to b
        b : ndarray(d)
            A d-dimensional vector that should end up residing at
            the north pole (0,0,...,0,1)

        """
        if (len(a.shape) > 1 and np.min(a.shape) > 1) or (
            len(b.shape) > 1 and np.min(b.shape) > 1
        ):
            print("Error: a and b need to be 1D vectors")
            return None
        a = a.flatten()
        a = a / np.sqrt(np.sum(a**2))
        d = a.size

        if b.size == 0:
            b = np.zeros(d)
            b[-1] = 1
        b = b / np.sqrt(np.sum(b**2))

        c = a - np.sum(b * a) * b
        # If a numerically coincides with b, don't rotate at all
        if np.sqrt(np.sum(c**2)) < 1e-15:
            return np.eye(d)

        # Otherwise, compute rotation matrix
        c = c / np.sqrt(np.sum(c**2))
        lam = np.sum(b * a)
        beta = np.sqrt(1 - np.abs(lam) ** 2)
        rot = (
            np.eye(d)
            - (1 - lam) * (c[:, None].dot(c[None, :]))
            - (1 - lam) * (b[:, None].dot(b[None, :]))
            + beta * (b[:, None].dot(c[None, :]) - c[:, None].dot(b[None, :]))
        )
        return rot


class LensMapUtils:
    """
    Utilities for manipulating lens space-valued maps.

    """

    @staticmethod
    def lens_3D_to_disk_3D(X, q):
        """
        Project points on the unit sphere in C^2 representing points in the 3-dimensional
        lens space corresponding to the prime q to points on the unit disk in R^3.

        Parameters
        ----------
        X : complex ndarray(n,2)
            Points on the unit sphere in C^2.

        q : int
            Prime number such that X represents points in the 3-dimensional lens space
            corresponding to this prime.

        Returns
        -------
        real ndarray(n,3)
            Points on the unit disk in R^3.

        """

        def _point_lens_to_sphere(p, q):
            p1 = p[0]
            p2 = p[1]
            arg_z = np.mod(np.angle(p1), 2 * np.pi)
            theta = np.mod(arg_z, 2 * np.pi / q)

            k = np.floor((arg_z - theta) / (2 * np.pi / q))

            phi = np.mod(np.angle(p2), 2 * np.pi) - 2 * k * np.pi / q

            r = np.abs(p2)
            x, y, z = (
                r * np.cos(phi),
                r * np.sin(phi),
                (q / np.pi) * (theta - np.pi / q) * np.sqrt(1 - r**2),
            )
            return [x, y, z]

        return np.array([_point_lens_to_sphere(p, q) for p in X])



import numpy as np
from .utils import PartUnity, EquivariantPCA
from .emcoords import EMCoords


class LensCoords(EMCoords):
    """
    Object that performs multiscale lens coordinates via
    persistent cohomology of sparse filtrations (Polanco, Perea 2019).

    Parameters
    ----------
    X: ndarray(N, d)
        A point cloud with N points in d dimensions
    n_landmarks: int
        Number of landmarks to use
    distance_matrix: boolean
        If true, treat X as a distance matrix instead of a point cloud
    prime : int
        Field coefficient with which to compute rips on landmarks
    maxdim : int
        Maximum dimension of homology. Only dimension 1 is needed for lens coordinates,
        but it may be of interest to see other dimensions (e.g. for a torus)
    partunity_fn: ndarray(n_landmarks, N) -> ndarray(n_landmarks, N)
        A partition of unity function

    """

    def __init__(
        self, X, n_landmarks, distance_matrix=False, prime=3, maxdim=1, verbose=False
    ):
        EMCoords.__init__(
            self,
            X=X,
            n_landmarks=n_landmarks,
            distance_matrix=distance_matrix,
            prime=prime,
            maxdim=maxdim,
            verbose=verbose,
        )
        self.type_ = "lens"

    def get_coordinates(
        self,
        perc=0.9,
        cocycle_idx=0,
        complex_dim=2,
        partunity_fn=PartUnity.linear,
        standard_range=True,
        projective_dim_red_mode="exponential"
    ):
        """
        Get lens coordinates.

        Parameters
        ----------
        perc : float
            Percent coverage. Must be between 0 and 1.
        cocycle_idx : list
            Add the cocycles together, sorted from most to least persistent.
        complex_dim : integer
            Describes the dimension of the lens space. The lens space is obtained by
            taking a quotient of the unit sphere inside the complex plane to the power
            of complex_dimension.
        partunity_fn: (dist_land_data, r_cover) -> phi
            A function from the distances of each landmark to a bump function.
        standard_range : bool
            Whether to use the parameter perc to choose a filtration parameter that guarantees
            that the selected cohomology class represents a class in the Cech complex.
        projective_dim_red_mode : string
            Either "one-by-one", "exponential", or "direct". How to perform equivariant
            dimensionality reduction. "exponential" usually works best, being fast
            without compromising quality.

        Returns
        -------
        ndarray(N, complex_dim-1)
            The lens coordinates

        """

        n_landmarks = self.n_landmarks_

        homological_dimension = 1
        cohomdeath_rips, cohombirth_rips, cocycle = self.get_representative_cocycle(
            cocycle_idx, homological_dimension
        )

        r_cover, _ = EMCoords.get_cover_radius(
            self, perc, cohomdeath_rips, cohombirth_rips, standard_range
        )

        varphi, ball_indx = EMCoords.get_covering_partition(self, r_cover, partunity_fn)

        root_of_unity = np.exp(2 * np.pi * 1j / self.prime_)

        cocycle_matrix = np.ones( (n_landmarks, n_landmarks), np.complex_ )
        cocycle_matrix[cocycle[:, 0], cocycle[:, 1]] = root_of_unity ** cocycle[:, 2]
        cocycle_matrix[cocycle[:, 1], cocycle[:, 0]] = (1 / root_of_unity) ** cocycle[:, 2]

        class_map = np.sqrt(varphi.T) * cocycle_matrix[ball_indx[:], :]

        epca = EquivariantPCA.ppca(class_map, complex_dim-1, projective_dim_red_mode, self.verbose)
        self.variance_ = epca["variance"]

        return epca["X"]
